from typing import Any, Type
import msgspec
import numpy as np

from niess.bifrost import Tank as BIFROSTTank, Primary as BIFROSTPrimary

from niess.components import (
    DirectSecondary,
    IndirectSecondary,
    IdealCrystal,
    Crystal,
    Wire,
    DiscreteWire,
    DiscreteTube,
    He3Tube,
    Aperture,
    Jaw,
    Slit,
    Chopper,
    DiscChopper,
    FermiChopper,
    Collimator,
    SollerCollimator,
    RadialCollimator,
    Component,
    Attenuator,
    Filter,
    NCrystalFilter,
    OrderedFilter,
    RadialFilterCollimator,
    Guide,
    EllipticGuide,
    TaperedGuide,
    StraightGuide,
    StraightGuides,
    TaperedGuides,
    Moderator,
    FissionChamber,
    He3Monitor,
    BeamCurrentMonitor,
    GEM2D,
    ESSource,
    Section,
)

from .scipp import (
    SCIPP_MODEL_DECODE, SCIPP_MODEL_ENCODE, ScippModel, from_scipp_model,
    SCIPP_TO_DICT, DICT_TO_SCIPP
)
from .mccode import (
    MCCODE_MODEL_ENCODE, MCCODE_MODEL_DECODE, McCodeModel
)

MODEL_ENCODE = {
    BIFROSTTank: 'BIFROST Tank',
    BIFROSTPrimary: 'BIFROST Primary',
    DirectSecondary: 'DirectSecondary',
    IndirectSecondary: 'IndirectSecondary',
    IdealCrystal: 'IdealCrystal',
    Crystal: 'Crystal',
    Wire: 'Wire',
    DiscreteWire: 'DiscreteWire',
    DiscreteTube: 'DiscreteTube',
    He3Tube: 'He3Tube',
    Aperture: 'Aperture',
    Jaw: 'Jaw',
    Slit: 'Slit',
    Chopper: 'Chopper',
    DiscChopper: 'DiscChopper',
    FermiChopper: 'FermiChopper',
    Collimator: 'Collimator',
    SollerCollimator: 'SollerCollimator',
    RadialCollimator: 'RadialCollimator',
    Component: 'Component',
    Attenuator: 'Attenuator',
    Filter: 'Filter',
    NCrystalFilter: 'NCrystalFilter',
    OrderedFilter: 'OrderedFilter',
    RadialFilterCollimator: 'RadialFilterCollimator',
    Guide: 'Guide',
    EllipticGuide: 'EllipticGuide',
    TaperedGuide: 'TaperedGuide',
    StraightGuide: 'StraightGuide',
    StraightGuides: 'StraightGuides',
    TaperedGuides: 'TaperedGuides',
    Moderator: 'Moderator',
    FissionChamber: 'FissionChamber',
    He3Monitor: 'He3Monitor',
    BeamCurrentMonitor: 'BeamCurrentMonitor',
    GEM2D: 'GEM2D',
    ESSource: 'ESSource',
    Section: 'Section',
}
MODEL_DECODE = {v: k for k, v in MODEL_ENCODE.items()}


class Model(msgspec.Struct):
    name: str
    obj: msgspec.Raw

    @classmethod
    def from_value(cls, obj: Any, encoder=None):
        if encoder is None:
            raise ValueError("An encoder must be provided")
        obj_type = type(obj)
        if obj_type in SCIPP_TO_DICT:
            model_type = SCIPP_MODEL_ENCODE[obj_type]
            obj = SCIPP_TO_DICT[obj_type](obj)
        elif obj_type in MCCODE_MODEL_ENCODE:
            model_type = MCCODE_MODEL_ENCODE[obj_type]
            if hasattr(obj, 'to_dict'):
                obj = obj.to_dict()
        else:
            model_type = MODEL_ENCODE[obj_type]
            if hasattr(obj, 'to_dict'):
                obj = obj.to_dict()
        return cls(model_type, msgspec.Raw(encoder.encode(obj)))


def to_model(model_encoder, obj):
    return Model(model_encoder, obj)


def traverse_decoded(a):
    if isinstance(a, dict):
        return traverse_decoded_dict(a)
    return traverse_decoded_one(a)


def traverse_decoded_dict(d: dict):
    """Convert special dicts based on specified conversion routines or
    traverse through the provided standard dict"""
    if 'name' in d and 'obj' in d and len(d) == 2:
        # one of the models:
        if d['name'] in DICT_TO_SCIPP:
            return DICT_TO_SCIPP[d['name']](d['obj'])
        print(f"{d['name']} should be handled")
    # if 'name' in d and 'obj' in d and d['name'] in DICT_TO_MCCODE:
    #     return DICT_TO_MCCODE[d['name']](d['obj'])
    for k, v in d.items():
        d[k] = traverse_decoded(v)
    return d


def traverse_decoded_one(a):
    if isinstance(a, dict):
        return traverse_decoded_dict(a)
    elif isinstance(a, list):
        return [traverse_decoded(x) for x in a]
    elif isinstance(a, tuple):
        return tuple(traverse_decoded(x) for x in a)
    return a


def from_model(model_decoder, model):
    if model.name in MODEL_DECODE:
        data = model_decoder.decode(model.obj)
        if isinstance(data, dict):
            data = traverse_decoded_dict(data)
            return MODEL_DECODE[model.name].from_dict(data)
        return data
    elif model.name in SCIPP_MODEL_DECODE:
        data = model_decoder.decode(model.obj)
        if isinstance(data, dict):
            return DICT_TO_SCIPP[model.name](data)
        return data
    elif model.name in MCCODE_MODEL_DECODE:
        data = model_decoder.decode(model.obj)
        if isinstance(data, dict):
            return MCCODE_MODEL_DECODE[model.name].from_dict(data)
        return data
    raise ValueError(f"Model {model.name} is not supported")


def encode_hook(encoder):
    def hook(obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif type(obj) in SCIPP_MODEL_ENCODE:
            return ScippModel.from_value(obj, encoder)
        raise TypeError(f"Object of type {type(obj)} is not supported")
    return hook


def decode_hook(scipp_model_decoder):
    def hook(typ: Type, obj: Any):
        if typ is ScippModel:
            return from_scipp_model(scipp_model_decoder, obj)
        # A data type that's been encoded as a ScippModel doesn't know it?
        if typ in SCIPP_MODEL_ENCODE and isinstance(obj, dict) and 'obj' in obj:
            return DICT_TO_SCIPP[SCIPP_MODEL_ENCODE[typ]](obj['obj'])
        raise TypeError(f"Object of type {type(obj)} is not supported")
    return hook


def schema_hook(typ):
    from .scipp import SCIPP_JSON_SCHEMA
    if typ in SCIPP_JSON_SCHEMA:
        return SCIPP_JSON_SCHEMA[typ]()
    raise TypeError(f"Object of type {typ} is not supported")



# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT


def test_ideal_crystal():
    from niess import IdealCrystal
    from scipp import scalar, array, vector, dot, norm, isclose

    pos = vector([0, 0, 0.], unit='m')
    tau = vector([0, 0, 1.], unit='1/m')

    crystal = IdealCrystal(pos, tau)

    # without any extent, the triangulated ideal crystal is a circle in the plane
    # perpendicular to tau, centered at pos
    vertices, triangle = crystal.triangulate()
    assert all(dot(v, tau) == scalar(0.) for v in vertices)
    assert isclose(norm(vertices.sum(dim='vertices')), scalar(0., unit='m'))

    assert isclose(crystal.momentum, scalar(1., unit='1/m'))

    x = crystal.scattering_angle(wavenumber=tau)
    assert isclose(x, scalar(60.0, unit='deg').to(unit=x.unit))


def box_edge_check(vs, x, y, z):
    from scipp import scalar, norm, isclose
    assert isclose(norm(vs['vertices', 1] - vs['vertices', 0]), scalar(x, unit='m'))
    assert isclose(norm(vs['vertices', 3] - vs['vertices', 2]), scalar(x, unit='m'))
    assert isclose(norm(vs['vertices', 5] - vs['vertices', 4]), scalar(x, unit='m'))
    assert isclose(norm(vs['vertices', 6] - vs['vertices', 7]), scalar(x, unit='m'))

    assert isclose(norm(vs['vertices', 2] - vs['vertices', 1]), scalar(y, unit='m'))
    assert isclose(norm(vs['vertices', 3] - vs['vertices', 0]), scalar(y, unit='m'))
    assert isclose(norm(vs['vertices', 6] - vs['vertices', 5]), scalar(y, unit='m'))
    assert isclose(norm(vs['vertices', 7] - vs['vertices', 4]), scalar(y, unit='m'))

    assert isclose(norm(vs['vertices', 4] - vs['vertices', 0]), scalar(z, unit='m'))
    assert isclose(norm(vs['vertices', 5] - vs['vertices', 1]), scalar(z, unit='m'))
    assert isclose(norm(vs['vertices', 6] - vs['vertices', 2]), scalar(z, unit='m'))
    assert isclose(norm(vs['vertices', 7] - vs['vertices', 3]), scalar(z, unit='m'))


def extrema_check(crystal, basis: list, extents: list):
    from scipp import concat, isclose
    box = crystal.bounding_box(basis=concat(basis, dim='basis'))
    lengths = box['limits', 1] - box['limits', 0]
    for index, extent in enumerate(extents):
        if extent is not None:
            assert isclose(lengths['basis', index], extent)


def test_crystal():
    from itertools import permutations
    from niess import Crystal
    from scipp import scalar, vector
    from scipp.spatial import rotations_from_rotvecs

    x, y, z = 1., 2., 0.01
    pos = vector([0, 0, 0.], unit='m')
    tau = vector([0, 0, 1.], unit='1/m')
    shape = vector([x, y, z], unit='m')
    orient = rotations_from_rotvecs(vector([0, 0, 0.], unit='deg'))
    mosaic = scalar(10., unit='arcminutes')

    crystal = Crystal(pos, tau, shape, orient, mosaic)

    vs, triangles = crystal.triangulate()
    expected = [[-x/2, -y/2, -z/2], [x/2, -y/2, -z/2], [x/2, y/2, -z/2], [-x/2, y/2, -z/2],
                [-x/2, -y/2, z/2], [x/2, -y/2, z/2], [x/2, y/2, z/2], [-x/2, y/2, z/2],]
    for index, v in enumerate(expected):
        assert all(a == b for a, b in zip(v, vs['vertices', index].values))
    box_edge_check(vs, x, y, z)

    v = [vector(vv, unit='m') for vv in ([1, 0, 0.], [0, 1, 0.], [0, 0, 1.])]
    s = [scalar(xx, unit='m') for xx in (x, y, z)]
    for p in permutations(range(3), 3):
        extrema_check(crystal, [v[i] for i in p], [s[i] for i in p])


def test_rotated_crystal():
    from itertools import permutations
    from niess import Crystal
    from scipp import scalar, vector, cos, sin
    from scipp.spatial import rotations_from_rotvecs

    x, y, z, angle = 1., 2., 0.01, 25.0
    pos = vector([0, 0, 0.], unit='m')
    tau = vector([0, 0, 1.], unit='1/m')
    shape = vector([x, y, z], unit='m')
    orient = rotations_from_rotvecs(vector([0, angle, 0.], unit='deg'))
    mosaic = scalar(10., unit='arcminutes')

    crystal = Crystal(pos, tau, shape, orient, mosaic)

    vs, triangles = crystal.triangulate()
    box_edge_check(vs, x, y, z)

    c, s = [fn(scalar(angle, unit='deg')) for fn in (cos, sin)]
    xz = scalar(x, unit='m') * c + scalar(z, unit='m') * s
    zx = scalar(z, unit='m') * c + scalar(x, unit='m') * s

    v = [vector(vv, unit='m') for vv in ([1, 0, 0.], [0, 1, 0.], [0, 0, 1.])]
    s = xz, scalar(y, unit='m'), zx
    for p in permutations(range(3), 3):
        extrema_check(crystal, [v[i] for i in p], [s[i] for i in p])





from scipp import Variable

def lengths_angle_circle(l0, l1, two_theta):
    from numpy import sqrt, cos, sin
    twice_sin_two_theta = 2 * sin(two_theta)
    radius = sqrt(l0**2 + l1**2 + l0 * l1 * cos(two_theta)) / twice_sin_two_theta
    z0 = l0 / 2
    x0 = (l0 * cos(2 * two_theta) - l1 * cos(two_theta)) / twice_sin_two_theta
    return x0, z0, radius


def three_point_circle(p0: Variable, p1: Variable, p2: Variable):
    from scipp import sqrt, dot, cross, scalar, isclose
    # vectors along each side of the triangle
    va, vb, vc = p1 - p0, p2 - p0, p2 - p1
    # and their lengths
    a, b, c = [sqrt(dot(x, x)) for x in (va, vb, vc)]
    # the semi-perimeter, s, defines the area sqrt(s(s-a)...)
    s = (a+b+c)/2
    # and the circumscribing circle has radius:
    r = a*b*c / (4*sqrt(s*(s-a)*(s-b)*(s-c)))
    # to find the center we must define the plane normal
    n = cross(vb, va) / a / b
    if sqrt(dot(n, n)) == scalar(0., unit=n.unit):
        raise RuntimeError("Points are collinear")
    n /= sqrt(dot(n, n))
    # the bisector of any one of the sides points to the center
    perp = [p / sqrt(dot(p,p)) for v in (-va, vb, -vc) for p in (cross(n, v),)]
    # the length of the bisector is given by the pythagorean theorem
    h = [sqrt(r*r - i*i/4) * p for p, i in zip(perp, (a, b, c))]
    # but whether each h should be positive or negative is determined by the distance
    # to the other two points
    z = []
    for x0, y0, z0, v, bi in zip((p0, p0, p1), (p1, p1, p2), (p2, p2, p0), (va, vb, vc), h):
      t = x0 + 0.5*v + bi
      if isclose(dot(t-y0, t-y0), r*r) and isclose(dot(t-z0, t-z0), r*r):
        z.append(t)
      else:
        z.append(x0 + 0.5*v - bi)

    # all three z points should be the same:
    if not isclose(z[0], z[1]) or not isclose(z[1], z[2]):
      raise RuntimeError(f"Mismatched central points,\n{p0=}\n{p1=}\n{p2=}\n{z = }")
    if z[0] != z[1] or z[1] != z[2]:
        # Only find the average of the three points if they are not identical to avoid division-by-three weirdness
        return (z[0]+z[1]+z[2])/3, r, n
    return z[0], r, n


def angle_between(a: Variable, b: Variable):
    from scipp import acos, dot, sqrt
    return acos(dot(a, b) / sqrt(dot(a, a)) / sqrt(dot(b, b)))


def rowland_blade_angles(beta: Variable, radius: Variable, count: int, width: Variable, gap=None):
    """
    Find the rotation angle needed for each crystal blade in a Rowland geometry

    :param beta:
        The angle subtended by the crystal array when viewed from the central axis of
        the Rowland cylinder
    :param radius:
        The radius of the Rowland cylinder
    :param count:
        The number of crystals in the array
    :param width:
        The width along the Rowland circumferential direction for any one crystal
    :param gap:
        The nominal spacing between crystals along the cylinder circumference
    :return:
    """
    # the crystals cover from (-beta/2, beta/2) *around the Rowland circle center point*, rho
    #   -beta/2                                        beta/2
    #  ----v------.------.------.------.------.------.---v----> rho
    #      |xxx|  |xxx|  |xxx|  |xxx|  |xxx|  |xxx|  |xxx|
    # So the angular range is broken up into N blade-width and (N-1) gap-width segments
    # The blade width is rho_blade ~= width / rowland_radius, so the gap width is given by
    #       rho_gap = (beta - N * rho_blade) / (N - 1)
    from scipp import atan2, isclose, scalar, arange
    if gap is None:
        r_width = 2 * atan2(y=0.5 * width, x=radius.to(unit=width.unit))
        r_gap = (beta - count * r_width) / (count - 1)
    else:
        # Follow the RTP method to calculate radial 'width' of each blade -- which might be greater than actual width
        from numpy import pi
        r_gap = gap / (2 * pi * radius.to(unit=gap.unit)) * scalar(2 * pi, unit='radian')
        # r_gap = 2 * atan2(y=0.5 * gap, x=radius.to(unit=gap.unit))
        r_width = (beta - (count - 1) * r_gap) / count

    half_count = count >> 1
    angles = (r_width + r_gap) * arange(start=-half_count, stop=half_count+1, dim='blade')
    if not isclose(angles['blade', half_count], scalar(0., unit='radian')):
        print(f"{beta = }\n{radius = }\n{count = }\n{width = }\n{r_width = }\n{r_gap = }\n{angles = :c}")
        raise RuntimeError(f"Central angle should be zero but is {angles[count>>1]}.")
    if gap is None and not isclose(angles[-1], beta/2 - 0.5 * r_width):
        raise RuntimeError("Last angle should be half a radial-width from the maximum coverage angle!")
    return angles


def rowland_blades(
        source: Variable,
        position: Variable,
        focus: Variable,
        alpha: Variable,
        width: Variable,
        count: int,
        tau: Variable,
        gap=None
):
    """
    Find the positions and normal vectors for the crystals arranged on the Rowland
    geometry cylinder connecting `source`, `position` and `focus`

    :param source:
        The position of the point-source from which all neutrons arrive
    :param position:
        The central on-cylinder position of this device; this is not the center of
        mass for the crystals since they are on the surface of the cylinder.
    :param focus:
        The position of the point-focus to which all neutrons are directed
    :param alpha:
        *Half* of the angular width of the crystals when viewed from the source
        (or focus) the angular size of the array viewed from the Rowland cylinder centre
        is exactly four times this value.
    :param width:
        The size of each crystal comprising the array, along the circumferential
        direction of the Rowland cylinder
    :param count:
        The number of crystals positioned along the Rowland cylinder circumference
    :param tau:
        The Q value of the reflection used by these crystals (2 pi / d_spacing)
    :param gap:
        The proposed spacing between successive crystals in the array
    :return:
    """
    from scipp.spatial import rotations_from_rotvecs

    center, radius, normal = three_point_circle(source, position, focus)
    angles = rowland_blade_angles(4.0 * alpha, radius, count, width, gap)

    # for each angle, create the rotation matrix needed to rotate (position - center)
    rotations = rotations_from_rotvecs(rotation_vectors=angles * normal)

    # *broadcast* the rotations to the center-to-analyzer-position vector
    blade_positions = rotations * (position - center) + center
    assert blade_positions.shape == rotations.shape

    # find the crystal-normal directions; the central one is set by Bragg's law, while the remaining ones
    # are rotated by half their Rowland angles
    rotations = rotations_from_rotvecs(rotation_vectors=0.5 * angles * normal)
    blade_taus = rotations * tau

    return blade_positions, blade_taus



from __future__ import annotations

from scipp import Variable
from .component import Component
from mccode_antlr.assembler import Assembler
from mccode_antlr.instr import Instance


class Filter(Component):
    composition: str | None
    temperature: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import scalar as s, vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        composition = cal.get('composition')
        temperature = cal.get('temperature')
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='deg')))
        return cls(
            name=name,
            composition=composition,
            temperature=temperature,
            position=position,
            orientation=orientation
        )

    @property
    def cfg(self):
        if self.composition is None:
            return '""'
        if '.ncmat' in self.composition:
            sample = self.composition
        else:
            sample = f'{self.composition}.ncmat'
        if ';temp=' not in sample:
            sample = f'{sample};temp={self.temperature:c}'.replace(' ', '')
        return '"' + sample + '"'


class NCrystalFilter(Filter):
    """A powder or amorphous material that effects a beam with physics from NCrystal

    Likely useful NCrystal data sets include:
        AcrylicGlass_C5O2H8
        Al_sg225
        Be_sg194
        Polycarbonate_C16O3H14
        C_sg194_pyrolytic_graphite
    """
    width: Variable
    height: Variable
    length: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import scalar as s, vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        width = cal.get('width')
        height = cal.get('height')
        length = cal.get('length')
        composition = cal.get('composition')
        temperature = cal.get('temperature', s(300., unit='K'))
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0.,0, 0], unit='deg')))
        return cls(
            name=name,
            position=position,
            orientation=orientation,
            width=width,
            height=height,
            length=length,
            composition=composition,
            temperature=temperature
        )

    def __mccode__(self) -> tuple[str, dict]:
        params = dict()
        params['xwidth'] = self.width.to(unit='m').value
        params['yheight'] = self.height.to(unit='m').value
        params['zdepth'] = self.length.to(unit='m').value
        params['cfg'] = self.cfg
        return 'NCrystal_sample', params


class OrderedFilter(NCrystalFilter):
    """(likely) A Bragg scattering filter, e.g., Pyrolytic Graphite"""
    tau: Variable # the direction and lattice spacing used in the filter


class Attenuator(NCrystalFilter):
    """A pneumatically insertable absorbing or scattering material

    Likely materials
    | name                                             | NCrystal dataset         |
    |--------------------------------------------------|--------------------------|
    | Poly(methyl methacrylate)                        |  'AcrylicGlass_C5O2H8'   |
    |  AKA Plexiglass, Lucite, Perspex, acrylic, etc.  |                          |
    | Polycarbonate                                    | 'Polycarbonate_C16O3H14' |
    """
    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        from mccode_antlr.common import InstrumentParameter, Expr
        parameter = InstrumentParameter.parse(f"int {self.name}_in = 0")
        assembler.instrument.add_parameter(parameter, ignore_repeated=True)
        comp = super().to_mccode(assembler, at, rotate)
        comp.WHEN(Expr.parameter(parameter.name))
        return comp


class RadialFilterCollimator(Filter):
    """A combination radial collimator and filter, as used on, e.g., BIFROST

    Typically the fitler material should be beryllium, but other options may prove
    useful in the future.
    """
    height: Variable
    filter_inner_radius: Variable
    filter_outer_radius: Variable
    collimator_inner_radius: Variable
    collimator_outer_radius: Variable
    angle_width: Variable
    collimation_angle: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        height = cal.get('height')
        inner_radius = cal.get('inner_radius')
        outer_radius = cal.get('outer_radius')
        filter_inner_radius = cal.get('filter_inner_radius', inner_radius)
        filter_outer_radius = cal.get('filter_outer_radius', outer_radius)
        collimator_inner_radius = cal.get('collimator_inner_radius', inner_radius)
        collimator_outer_radius = cal.get('collimator_outer_radius', outer_radius)
        angle_width = cal.get('angle_width')
        channel_count = cal.get('channel_count', 1)
        collimation_angle = cal.get('collimation_angle', angle_width / channel_count)
        composition = cal.get('composition')
        temperature = cal.get('temperature')
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='deg')))
        return cls(
            name=name,
            height=height,
            filter_inner_radius=filter_inner_radius,
            filter_outer_radius=filter_outer_radius,
            collimator_inner_radius=collimator_inner_radius,
            collimator_outer_radius=collimator_outer_radius,
            angle_width=angle_width,
            collimation_angle=collimation_angle,
            composition=composition,
            temperature=temperature,
            position=position,
            orientation=orientation
        )

    def __mccode__(self) -> tuple[str, dict]:
        params = {
            "yheight": self.height.to(unit='m').value,
            "filter_minimum_radius": self.filter_inner_radius.to(unit='m').value,
            "filter_maximum_radius": self.filter_outer_radius.to(unit='m').value,
            "collimator_minimum_radius": self.collimator_inner_radius.to(unit='m').value,
            "collimator_maximum_radius": self.collimator_outer_radius.to(unit='m').value,
            "angle_width": self.angle_width.to(unit='deg').value,
            "collimation": self.collimation_angle.to(unit='deg').value,
            "cfg": self.cfg,
        }
        return 'Radial_filter_collimator', params

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        """Overload to ensure the registry we need is present"""
        from niess.mccode import ensure_registry
        ensure_registry(assembler, 'mcdotstar/mcstas-radial-filter-collimator@main')
        comp = super().to_mccode(assembler, at, rotate)
        return comp


def make_aluminum(name, position, orientation, width, height, length):
    from scipp import scalar
    return NCrystalFilter(
        name=name, position=position, orientation=orientation,
        width=width, height=height, length=length,
        composition='Al_sg225', temperature=scalar(300, unit='K')
    )



from dataclasses import dataclass, field
from typing import List, Tuple, Union
from scipp import Variable, vector, zeros

from .detectors import DiscreteTube
from .crystals import IdealCrystal


def serialize_to(g, name, what):
    from numpy import vstack
    ser = vstack([x.serialize() for x in what])
    g.create_dataset(name, data=ser)
    g[name].attrs['py_class'] = type(what[0]).__name__


@dataclass
class DirectSecondary:
    detectors: List[DiscreteTube]
    sample_at: Variable = field(default_factory=lambda: vector([0., 0., 0.], unit='m'))

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('DirectSecondary(...)')
        else:
            with p.group(4, 'DirectSecondary(', ')'):
                p.text(f'{len(self.detectors)} detectors')
                p.breakable()
                p.text(f'sample at {self.sample_at.value} {self.sample_at.unit}')


    def __eq__(self, other):
        return all(sd == od for sd, od in zip(self.detectors, other.detectors)) and self.sample_at == other.sample_at

    def approx(self, other):
        from scipp import allclose
        if not isinstance(other, DirectSecondary):
            return False
        return all(sd.approx(od) for sd, od in zip(self.detectors, other.detectors)) \
               and allclose(self.sample_at, other.sample_at)

    def __post_init__(self):
        from scipp import DType
        if any([not isinstance(x, DiscreteTube) for x in self.detectors]):
            raise RuntimeError("DiscreteTube detectors expected")
        if self.sample_at.dtype != DType.vector3:
            raise RuntimeError("sample_at should be a scipp.DType('vector3')")

    def final_vector(self, detector: int, element: int) -> Variable:
        p = self.detectors[detector].index_position(element)
        return p - self.sample_at.to(unit=p.unit)

    def final_distance(self, detector, element) -> Variable:
        from scipp import sqrt, dot
        v = (self.final_vector(detector, element))
        return sqrt(dot(v, v))

    def final_direction(self, detector, element) -> Variable:
        from scipp import sqrt, dot
        v = self.final_vector(detector, element)
        return v / sqrt(dot(v, v))


@dataclass
class IndirectSecondary:
    detectors: List[DiscreteTube]
    analyzers: List[IdealCrystal]
    analyzer_per_detector: List[int]
    sample_at: Variable = field(default_factory=lambda: vector([0., 0., 0.], unit='m'))
    analyzer_map: Variable = field(default_factory=lambda: zeros(shape=[0, 0], dims=['channel', 'pair']))
    detector_map: Variable = field(default_factory=lambda: zeros(shape=[0, 0, 0], dims=['channel', 'pair', 'tube']))

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('IndirectSecondary(...)')
        else:
            with p.group(4, 'IndirectSecondary(', ')'):
                p.text(f'{len(self.detectors)} detectors')
                p.breakable()
                p.text(f'{len(self.analyzers)} analyzers')
                p.breakable()
                p.text(f'sample at {self.sample_at.value} {self.sample_at.unit}')

    def __eq__(self, other):
        if not isinstance(other, IndirectSecondary):
            return False
        if self.sample_at != other.sample_at:
            return False
        if not all(sd == od for sd, od in zip(self.detectors, other.detectors)):
            return False
        if not all(sa == oa for sa, oa in zip(self.analyzers, other.analyzers)):
            return False
        return all(sad == oad for sad, oad in zip(self.analyzer_per_detector, other.analyzer_per_detector))

    def approx(self, other):
        from scipp import allclose
        if not isinstance(other, IndirectSecondary):
            return False
        if not allclose(self.sample_at, other.sample_at):
            return False
        if not all(sd.approx(od) for sd, od in zip(self.detectors, other.detectors)):
            return False
        if not all(sa.approx(oa) for sa, oa in zip(self.analyzers, other.analyzers)):
            return False
        return all(sad == oad for sad, oad in zip(self.analyzer_per_detector, other.analyzer_per_detector))

    def __post_init__(self):
        from scipp import DType
        n_det = len(self.detectors)
        n_ana = len(self.analyzers)
        if len(self.analyzer_per_detector) != n_det:
            raise RuntimeError("The detector-to-analyzer map must have one entry per detector")
        if any((x < 0 or x > n_ana for x in self.analyzer_per_detector)):
            raise RuntimeError("The analyzer index for each detector must be valid")
        if any([not isinstance(x, DiscreteTube) for x in self.detectors]):
            raise RuntimeError("DiscreteTube detectors expected")
        if any([not isinstance(x, IdealCrystal) for x in self.analyzers]):
            raise RuntimeError("IdealCrystal analyzers expected")
        if self.sample_at.dtype != DType.vector3:
            raise RuntimeError("sample_at should be a scipp.DType('vector3')")

    def scattering_plane_normal(self, analyzer) -> Variable:
        from scipp import sqrt, dot, cross
        tau = self.analyzers[analyzer].tau  # points into the scattering plane
        mod_q = sqrt(dot(tau, tau))

        a = self._analyzer_center(analyzer)
        mod_a = sqrt(dot(a, a))

        # the cross product: tau x a
        n = cross(tau / mod_q, a / mod_a)
        return n / sqrt(dot(n, n))

    def _analyzer_center(self, analyzer) -> Variable:
        a = self.analyzers[analyzer].position
        return a - self.sample_at.to(unit=a.unit)

    def _detector_partial_vectors(self, detector: int, d: Variable) \
            -> Tuple[Variable, Variable, Variable, Variable, Variable, Variable, Variable, Variable]:
        from scipp import dot, sqrt
        # TODO CONTINUE!

        analyzer = self.analyzer_per_detector[detector]
        a = self._analyzer_center(analyzer)
        n = self.scattering_plane_normal(analyzer)

        d_dot_n = dot(d, n)
        d_n = d_dot_n * n

        # the vector from the analyzer center to in-plane detector position is the scattered wavevector direction
        f = d - d_n - a

        mod_a = sqrt(dot(a, a))
        mod_d_n = sqrt(dot(d_n, d_n))
        mod_f = sqrt(dot(f, f))

        mod_a_n = (mod_a / (mod_a + mod_f)) * mod_d_n
        a_n = mod_a_n * n

        return a, f, d_n, a_n, mod_a, mod_f, mod_d_n, mod_a_n

    def detector_vector(self, detector, element) -> Variable:
        p = self.detectors[detector].index_position(element)
        return p - self.sample_at.to(unit=p.unit)

    def continuous_detector_vector(self, detector, ratio) -> Variable:
        p = self.detectors[detector].continuous_position(ratio)
        return p - self.sample_at.to(unit=p.unit)

    def _analyzer_vector(self, detector, d) -> Variable:
        a, _, _, a_n, _, _, _, _ = self._detector_partial_vectors(detector, d)
        return a + a_n

    def analyzer_vector(self, detector, element) -> Variable:
        return self._analyzer_vector(detector, self.detector_vector(detector, element))

    def continuous_analyzer_vector(self, detector, ratio) -> Variable:
        return self._analyzer_vector(detector, self.continuous_detector_vector(detector, ratio))

    def final_direction(self, detector, element) -> Variable:
        from scipp import sqrt, dot
        av = self.analyzer_vector(detector, element)
        a = sqrt(dot(av, av))
        return av / a

    def continuous_final_direction(self, detector, ratio) -> Variable:
        from scipp import sqrt, dot
        av = self.continuous_analyzer_vector(detector, ratio)
        return av / sqrt(dot(av, av))

    def _final_distance(self, detector, d: Variable) -> Variable:
        from scipp import sqrt
        _, _, _, _, la, lf, ld_n, _ = self._detector_partial_vectors(detector, d)
        return sqrt((la + lf) * (la + lf) + ld_n * ld_n)

    def final_distance(self, detector, element) -> Variable:
        return self._final_distance(detector, self.detector_vector(detector, element))

    def continuous_final_distance(self, detector, ratio) -> Variable:
        return self._final_distance(detector, self.continuous_detector_vector(detector, ratio))

    def _broadcast_continuous_per_tube(self, detector_index: Variable):
        from scipp import concat, dot, sqrt
        dim = detector_index.dims[0]

        # print(detector_index)
        # detectors = detector_index.transform_coords('d', graph={'d': lambda x: self.detectors[x]})
        # print(detectors)
        detectors = [self.detectors[i] for i in detector_index.values]

        at = concat([d.at for d in detectors], dim=dim)
        to = concat([d.to for d in detectors], dim=dim)

        sd = (at + to) / 2.0 - self.sample_at

        analyzers = [self.analyzer_per_detector[i] for i in detector_index.values]
        sac = concat([self._analyzer_center(analyzer) for analyzer in analyzers], dim=dim)
        n = concat([self.scattering_plane_normal(analyzer) for analyzer in analyzers], dim=dim)

        # sample-to-detection-point vector component which is *out of the nominal scattering plane*
        # # This *SHOULD BE* zero for the tube-centre, unless if the tube is off centre
        d_dot_n = dot(sd, n)

        # the vector from the analyzer center to in-plane detector position is the scattered wave-vector direction
        f = sd - d_dot_n * n - sac

        mod_a = sqrt(dot(sac, sac))
        mod_f = sqrt(dot(f, f))

        mod_a_n = (mod_a / (mod_a + mod_f)) * d_dot_n
        a_n = mod_a_n * n

        # sample to analyzer interaction-point vector
        sa = sac + a_n
        # analyzer interaction-point to detection-point vector
        ad = sd - sa
        return {'sample_analyzer': sa, 'analyzer_detector': ad, 'sample_analyzer_centre': sac,
                'analyzer_detector_centre': f,
                'length_sample_analyzer_centre': mod_a, 'length_analyzer_detector_centre': mod_f,
                'signed_length_detector_position': d_dot_n}

    def _broadcast_continuous_common(self, detector_index: Variable, ratio: Variable):
        from scipp import concat, scalar, dot, sqrt, acos
        dim = detector_index.dims[0]

        detectors = [self.detectors[i] for i in detector_index.values]

        at = concat([d.at for d in detectors], dim=dim)
        to = concat([d.to for d in detectors], dim=dim)

        sd = (scalar(1) - ratio) * at + ratio * to - self.sample_at

        analyzers = [self.analyzer_per_detector[i] for i in detector_index.values]
        sac = concat([self._analyzer_center(analyzer) for analyzer in analyzers], dim=dim)
        n = concat([self.scattering_plane_normal(analyzer) for analyzer in analyzers], dim=dim)

        # sample-to-detection-point vector component which is *out of the nominal scattering plane*
        d_dot_n = dot(sd, n)

        # the vector from the analyzer center to in-plane detector position is the scattered wave-vector direction
        f = sd - d_dot_n * n - sac

        mod_a = sqrt(dot(sac, sac))
        mod_f = sqrt(dot(f, f))

        mod_a_n = (mod_a / (mod_a + mod_f)) * d_dot_n
        a_n = mod_a_n * n

        # sample to analyzer interaction-point vector
        sa = sac + a_n
        # analyzer interaction-point to detection-point vector
        ad = sd - sa
        return {'sample_analyzer': sa, 'analyzer_detector': ad, 'sample_analyzer_centre': sac,
                'analyzer_detector_centre': f, 'scattering_plane_normal': n,
                'length_sample_analyzer_centre': mod_a, 'length_analyzer_detector_centre': mod_f,
                'signed_length_detector_position': d_dot_n}

    def broadcast_continuous_theta(self, detector_index: Variable, ratio: Variable):
        from scipp import scalar, dot, sqrt, acos
        # have an angle between them which is 2*theta
        values = self._broadcast_continuous_common(detector_index, ratio)
        # sample to analyzer interaction-point vector
        sa = values['sample_analyzer']
        # analyzer interaction-point to detection-point vector
        ad = values['analyzer_detector']
        two_theta = acos(dot(sa, ad) / sqrt(dot(sa, sa)) / sqrt(dot(ad, ad)))
        return two_theta / scalar(2)

    def broadcast_continuous_delta_a4(self, detector_index: Variable, ratio: Variable):
        from scipp import dot, sqrt, atan2
        values = self._broadcast_continuous_common(detector_index, ratio)
        sac, sa = values['sample_analyzer_centre'], values['sample_analyzer']
        n = values['scattering_plane_normal']

        c_hat = sac / sqrt(dot(sac, sac))
        a_hat = sa / sqrt(dot(sa, sa))

        ca = dot(c_hat, a_hat)
        na = dot(n, a_hat)
        delta_a4 = atan2(y=na, x=ca)

        return delta_a4

    def broadcast_continuous_final_distance(self, detector_index: Variable, ratio: Variable):
        from scipp import sqrt
        values = self._broadcast_continuous_common(detector_index, ratio)
        a = values['length_sample_analyzer_centre']
        f = values['length_analyzer_detector_centre']
        d = values['signed_length_detector_position']
        return sqrt((a + f) * (a + f) + d * d)

    def broadcast_continuous_analyzer_distance(self, detector_index: Variable):
        values = self._broadcast_continuous_per_tube(detector_index)
        return values['length_sample_analyzer_centre']

    def broadcast_continuous_detector_distance(self, detector_index: Variable):
        values = self._broadcast_continuous_per_tube(detector_index)
        return values['length_analyzer_detector_centre']

    def broadcast_continuous_a6(self, detector_index: Variable):
        from scipp import sqrt, dot, acos
        values = self._broadcast_continuous_per_tube(detector_index)
        sa = values['sample_analyzer_centre']
        ad = values['analyzer_detector_centre']
        return acos(dot(sa, ad) / sqrt(dot(sa, sa)) / sqrt(dot(ad, ad)))

    def broadcast_continuous_plane_spacing(self, detector_index: Variable):
        from scipp import concat, scalar, dot, sqrt, acos
        dim = detector_index.dims[0]
        analyzers = [self.analyzer_per_detector[i] for i in detector_index.values]
        d = concat([self.analyzers[analyzer].plane_spacing for analyzer in analyzers], dim=dim)
        return d



# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT
from .components import (
    DirectSecondary,
    IndirectSecondary,
    IdealCrystal,
    Crystal,
    Wire,
    DiscreteWire,
    DiscreteTube,
    He3Tube
)

__all__ = [
DirectSecondary,
    IndirectSecondary,
    IdealCrystal,
    Crystal,
    Wire,
    DiscreteWire,
    DiscreteTube,
    He3Tube
]


from __future__ import annotations

from scipp import Variable


def __is_vector__(x: Variable):
    from scipp import DType
    return x.dtype == DType.vector3


def is_scipp_vector(v: Variable, name: str):
    from scipp import DType
    if v.dtype != DType.vector3:
        raise RuntimeError(f"The {name} must be a scipp.DType('vector3')")


def __is_quaternion__(x: Variable):
    from scipp import DType
    return x.dtype == DType.rotation3


def vector_to_vector_quaternion(fr: Variable, to: Variable):
    if not __is_vector__(fr) or not __is_vector__(to):
        raise RuntimeError("Two vectors required!")
    from scipp import sqrt, dot, cross
    from scipp.spatial import rotations
    from numpy import concatenate, expand_dims
    # following http://lolengine.net/blog/2013/09/18/beautiful-maths-quaternion-from-vectors
    u = fr / sqrt(dot(fr, fr))
    v = to / sqrt(dot(to, to))
    scalar_part = 0.5 * sqrt(2 + 2 * dot(u, v))
    vector_part = 0.5 * cross(u, v) / scalar_part
    values = concatenate((vector_part.values, expand_dims(scalar_part.values, axis=-1)), axis=-1)
    dims = vector_part.dims

    q = rotations(values=values, dims=dims)
    return q


def combine_triangulations(vts: list[tuple[Variable, list[list[int]]]]):
    from scipp import concat
    from numpy import cumsum, hstack
    if any((v.ndim != 1 for v, t in vts)):
        raise RuntimeError("All vertices expected to be 1-D lists of vectors")
    vdims = [v.dims[0] for v, t in vts]
    if any((d != vdims[0] for d in vdims)):
        raise RuntimeError("All vertex arrays expected to have the same dimension name")
    vdim = vdims[0]

    lens = [len(v) for v, t in vts]
    offset = hstack((0, cumsum(lens)))[:-1]
    faces = [[off + i for i in t] for off, (v, ts) in zip(offset, vts) for t in ts]

    vertices = concat([v for v, t in vts], dim=vdim)

    return vertices, faces


def write_off_file(vertices, faces, filename):
    stream = f"OFF\n{len(vertices)} {len(faces)} 0\n"
    for v in vertices.values:
        s = " ".join([f"{x:3.9f}" for x in v])
        stream += s + "\n"
    for v in faces:
        s = " ".join([f"{x}" for x in v])
        stream += f"{len(v)} {s}\n"
    with open(filename, 'w') as f:
        f.write(stream)


def combine_extremes(vs: list[Variable], horizontal: Variable, vertical: Variable):
    from scipp import concat, dot, sqrt, scalar, isclose, cross
    from numpy import argmax, argmin, hstack, unique
    is_scipp_vector(horizontal, 'horizontal')
    is_scipp_vector(vertical, 'vertical')
    map(lambda p: is_scipp_vector(p, 'x'), vs)
    if any((v.ndim != 1 for v in vs)):
        raise RuntimeError("All vertices expected to be 1-D lists of vectors")
    dim = vs[0].dims[0]
    if any((v.dims[0] != dim for v in vs)):
        raise RuntimeError("All vertex arrays expected to have the same dimension name")
    vs = concat(vs, dim)
    y = horizontal / sqrt(dot(horizontal, horizontal))
    z = vertical / sqrt(dot(vertical, vertical))
    if not isclose(dot(z, y), scalar(0.)):
        z = z - dot(z, y) * y
        z = z / sqrt(dot(z, z))
    x = cross(y, z)
    v_yz = vs - dot(x, vs) * x
    v_yz_pp = dot(v_yz, y + z).values
    v_yz_pm = dot(v_yz, y - z).values
    v_yz_p0 = dot(v_yz, y).values
    v_yz_0p = dot(v_yz, z).values
    max_pp, max_pm, max_p0, max_0p = [argmax(x) for x in (v_yz_pp, v_yz_pm, v_yz_p0, v_yz_0p)]
    min_pp, min_pm, min_p0, min_0p = [argmin(x) for x in (v_yz_pp, v_yz_pm, v_yz_p0, v_yz_0p)]

    idxs = hstack([max_p0, max_pp, max_0p, min_pm, min_p0, min_pp, min_0p, max_pm])  # order is important
    _, unique_index = unique(idxs, return_index=True)
    idxs = idxs[sorted(unique_index)]  # only indexes which are unique, in the same order as provided in idxs

    return vs[idxs]


def perpendicular_directions(direction: Variable):
    from scipp import sqrt, dot, cross, scalar, isclose, abs, vector
    is_scipp_vector(direction, 'direction')

    direction /= sqrt(dot(direction, direction))

    horizontal = vector([1., 0, 0]) if isclose(abs(direction.fields.y), scalar(1.)) else vector([0, -1., 0])
    horizontal -= dot(horizontal, direction) * direction
    horizontal /= sqrt(dot(horizontal, horizontal))

    vertical = cross(horizontal, direction)

    return horizontal, vertical


def bounding_box(points: Variable, basis: Variable):
    """ Find the minimal box with axes along `basis` that contains all `points`

    :param points:
        3-D vector points representing a continuous object in space
    :param basis:
        One or more basis vectors along which to calculate the bounding box.
        The basis _should_ be orthogonal, normal, and span space; but a projected
        basis is allowed (less than 3 basis vectors), normalized vectors will be
        enforced, and orthogonality is not checked.
    :return:
        The (2, len(basis)) limiting corners of the bounding box ordered from
        the minimum corner to the maximum corner; with named axis 'limits'
    """
    from scipp import norm, dot, min, max, concat
    is_scipp_vector(points, 'points')
    is_scipp_vector(basis, 'basis')
    basis = basis / norm(basis)
    if basis.ndim > 1:
        basis = basis.flatten(to='basis')

    projections = dot(points.flatten(to='points'), basis)
    lower = min(projections, dim='points')
    upper = max(projections, dim='points')
    box = concat((lower, upper), dim='limits')
    return box.transpose(['limits', basis.dim]) if basis.ndim > 1 else box


def combine_bounding_boxes(boxes: Variable, limits: str | None = None, basis: str | None = None):
    """Find the box which bounds one or more bounding boxes

    A bounding box can be represented as its minimum and maximum corners.
    This method leverages named `scipp.Variable` axes to identify the corners-axis.
    To allow dimensional flexibility, the box-axes are handled separately with the
    `scipp.Variable` axis representing them identified as the basis dimension.
    Any additional axis or axes are flattened and reduced to find the single
    all-containing bounding box.

    Params
    ------
    boxes:
        A (N, 2, ...) [or any permutation] array of bounding box limits.
    limits:
        The name of the array dimension that identifies the minimum and maximum corner.
        This dimension must be exactly 2-elements long.
        Default = 'limits'
    basis:
        The name of the array dimension that identifies the box basis, (e_1, e_2, ...).
        The length of the array in this dimension is the dimensionality of the space.
        Default = 'basis'
    Returns
    -------
    :
        The (2, N) all-encompassing bounding box, with limit-dimension and
        basis-dimension names matching those in the input bound boxes array
    """
    from scipp import min, max, concat
    limits = limits or 'limits'
    basis = basis or 'basis'
    if limits not in boxes.dims or 2 != boxes.sizes[limits]:
        raise ValueError(f'The boxes must have a `limits` ({limits}) axis of length 2')
    if basis not in boxes.dims:
        raise ValueError(f'The boxes must have a `basis` ({basis}) axis')

    others = [dim for dim in boxes.dims if dim not in (limits, basis)]
    if len(others):
        from uuid import uuid4
        dim = str(uuid4())
        box = boxes.transpose([limits, basis, *others]).flatten(dims=others, to=dim)
        lower = min(box[limits, 0], dim=dim)
        upper = max(box[limits, 1], dim=dim)
        box = concat((lower, upper), dim=limits)
    else:
        box = boxes

    return box.transpose([limits, basis])


def mccode_ordered_angles(orientation: Variable):
    """Determine the McCode ordered Euler angles that represent a quaternion

    In McCode instruments, a rotation is specified as (x, y, z), indicating the
    ordered rotation, R_z(z) R_y(y) R_x(x), which is applied on the left of a vector,
        v_rotated = R_z(z) R_y(y) R_x(x) v

    Each R matrix is the 'standard' rotation matrix with the noted axis constant,
              [1  0      0     ]          [ cos(y) 0 sin(y)]          [cos(z) -sin(z) 0]
        R_x = [0 cos(x) -sin(x)]    R_y = [ 0      1 0     ]    R_z = [sin(z)  cos(z) 0]
              [0 sin(x)  cos(x)]          [-sin(y) 0 cos(y)]          [0       0      1]

    Parameters
    ----------
    orientation: `scipp.Variable`
        A quaternion-valued scalar representing the transformation between two
        orthonormal axes systems.

    Returns
    -------
    : `tuple[float, float, float]`
        The x, y, and z angles needed to represent the same transformation in a McCode
        instrument component instance definition line, e.g., `ROTATED (x, y, z)`.
    """
    if not __is_quaternion__(orientation):
        raise ValueError(f"{orientation=} expected to be a scipp quaternion")
    from math import asin, atan2, pi

    # Follow the suggestions at Wikipedia.com, and the gimbal-lock avoidance
    # from EuclideanSpace.com
    # https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/

    x, y, z, w = orientation.value
    lock = x * y + z * w
    if lock > 0.4999:
        # gimbal lock with straight-up orientation
        pitch = pi / 2
        roll  = 2 * atan2(x, w)
        yaw = 0
    elif lock < -0.4999:
        # gimbal lock with straight-down orientation (pitch = -90)
        pitch = -pi / 2
        roll = -2 * atan2(x, w)
        yaw = 0
    else:
        roll = atan2(2 * (w * x + y * z), 1 - 2 * (x * x + y * y))
        pitch = asin(2 * (w * y - x * z))
        yaw = atan2(2 * (w * z + x * y), 1 - 2 * (y *y + z * z))

    return 180 / pi * roll, 180 / pi * pitch, 180 / pi * yaw


def mccode_quaternion(x, y, z):
    from scipp import vector
    from scipp.spatial import rotations_from_rotvecs as r
    rx = r(vector([x, 0, 0], unit='deg'))
    ry = r(vector([0, y, 0], unit='deg'))
    rz = r(vector([0, 0, z], unit='deg'))
    q = rz * ry * rx
    return q


def at_relative(relative_position: Variable, relative_orientation: Variable, position: Variable):
    if not __is_vector__(relative_position):
        raise ValueError('The position of the relative coordinate system must be a vector')
    if not __is_quaternion__(relative_orientation):
        raise ValueError('The orientation of the relative coordinate system must be a quaternion')
    if isinstance(position, Variable) and not __is_vector__(position):
        from scipp import vector
        print('Warning, implicit relative z-axis positioning')
        position = position * vector([0, 0, 1.0])

    if not __is_vector__(position):
        raise ValueError('The position in the relative coordinate system must be a vector')
    from numpy import eye
    from scipy.spatial.transform import Rotation
    from scipp.spatial import affine_transform
    a_mat = eye(4)
    a_mat[:3, :3] = Rotation(relative_orientation.value).as_matrix() # rotation part
    a_mat[:3, 3] = relative_position.value  # translation part
    a = affine_transform(value=a_mat, unit=relative_position.unit)
    return (a * position.to(unit=a.unit)).to(unit=position.unit)


def at_relative_dict(relative_dict: dict, position: Variable):
    if 'position' not in relative_dict:
        raise ValueError('The relative dictionary must have a position')
    if 'orientation' not in relative_dict:
        raise ValueError('The relative dictionary must have a orientation')
    return at_relative(relative_dict['position'], relative_dict['orientation'], position)


def rotate_relative(relative_orientation: Variable, orientation: Variable):
    if not __is_quaternion__(relative_orientation):
        raise ValueError('The orientation of the relative coordinate system must be a quaternion')
    if not __is_quaternion__(orientation):
        raise ValueError('The orientation relative to the relative coordinate system must be a quaternion')
    return orientation * relative_orientation


from __future__ import annotations

from niess.utilities import calibration
from niess.components import He3Monitor
from niess.components.component import Base

def _elastic_monitor_from_params(params):
    from scipp import vector
    from scipp.spatial import rotations_from_rotvecs
    from .parameters import tank_parameters
    tp = tank_parameters()
    def par_or(par):
        return tp[par] if par not in params else params[par]

    # There is a possibility that some or all necessary parameters are missing
    # but there are not always good defaults to provide in all cases. What do we do?
    distance = par_or('sample_elastic_monitor_distance')
    angle = par_or('tank_elastic_monitor_angle')
    y, z = vector([0, 1, 0]), vector([0, 0, 1])
    ori = rotations_from_rotvecs(y * angle)  # is this the monitor orientation too?

    cal = par_or('elastic_monitor')
    cal['name'] = cal.get('name', 'elastic_monitor')
    cal['position'] = cal.get('position', ori * (z * distance))
    cal['orientation'] = cal.get('orientation', ori)

    return He3Monitor.from_calibration(cal)


class Tank(Base):
    from scipp import Variable
    from networkx import DiGraph
    from .channel import Channel
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.instr import Instance

    channels: tuple[Channel, ...]
    monitor: He3Monitor

    @classmethod
    def from_dict(cls, data):
        from .channel import Channel
        cs = data['channels']
        if not hasattr(cs, '__len__'):
            raise ValueError('Channels must have length (probably 9)')
        cs = tuple(c if isinstance(c, Channel) else Channel.from_dict(c) for c in cs)
        mn = data['monitor']
        if not isinstance(mn, He3Monitor):
            mn = He3Monitor.from_dict(mn)
        return cls(cs, mn)

    @staticmethod
    @calibration
    def from_calibration(cal: dict):
        """Construct a Tank from a calibration dictionary.

        Parameters
        ----------
        cal: dict
            If empty, or if no `'channels'` entry is present, the values provided by
            `py::module::niess.bifrost.parameters::known_channel_params` are used.
            If 'channels' is present, it should contain a dictionary with per-channel
            parameters stored under a key 'channel_params', which is an integer keyed
            dictionary of variant-free parameters. The length of the channel_params
            dictionary should match the length of cal['channels']['angles'], a
            float-valued 1-D scipp.array of the channel angle relative to the tank
            centerline, defaulting to 9-channels at -40:10:40 degrees (inclusive)

        """
        from scipp import arange, linspace
        from .channel import Channel
        from .parameters import known_channel_params
        from niess.utilities import variant_parameters
        params = cal.get('channels', known_channel_params())
        variants = [{'variant': x} for x in ('s', 'm', 'l')]
        # The central a4 angle for each channel, relative to the reference tank angle
        angles = linspace('channel', -40, 40, 9, unit='degree', dtype='float')
        angles = params.get('angles', angles)
        # Assume the channel variants cycle through ('s', 'm', 'l') as in reality
        channel_params = {i: variants[i % 3] for i in range(angles.size)}
        # but this can be overridden by specifying an integer-keyed dictionary
        # with the parameters for each channel (and .pop removes it from params if present)
        channel_params = params.pop('channel_params', channel_params)

        # Which we might need to update with per-variant/constant parameters
        for val in channel_params.values():
            val.update(variant_parameters(val, params))

        channels = [Channel.from_calibration(angles[i], **channel_params[i]) for i in range(9)]
        return Tank(tuple(channels), _elastic_monitor_from_params(cal))

    @staticmethod
    def unique_from_calibration(**params):
        from scipp import array
        from .channel import Channel
        channel_params = [{'variant': x} for x in ('s', 'm', 'l')]
        channel_params = {i: channel_params[i % 3] for i in range(3)}
        # but this can be overridden by specifying an integer-keyed dictionary with the parameters for each channel
        channel_params = params.get('channel_params', channel_params)
        # The central a4 angle for each channel, relative to the reference tank angle
        angles = params.get('angles',
                            array(values=[-40, -30, -20, -10, 0, 10, 20, 30, 40.], unit='degree', dims=['channel']))

        channels = [Channel.from_calibration(angles[i], **channel_params[i]) for i in range(3)]
        return Tank(tuple(channels), _elastic_monitor_from_params(params))

    def to_secondary(self, **params):
        from scipp import vector
        from ..components import IndirectSecondary

        sample_at = params.get('sample', vector([0, 0, 0.], unit='m'))

        detectors = []
        analyzers = []
        a_per_d = []
        for channel in self.channels:
            for arm in channel.pairs:
                analyzers.append(arm.analyzer.central_blade)
                detectors.extend(arm.detector.tubes)
                a_per_d.extend([len(analyzers) - 1 for _ in arm.detector.tubes])

        from scipp import arange
        nc = len(self.channels)
        np = len(self.channels[0].pairs)
        a = arange(start=0, stop=len(analyzers), dim='n').fold('n', sizes={'channel': nc, 'pair': np})
        d = arange(start=0, stop=len(detectors), dim='n').fold('n', sizes={'channel': nc, 'pair': np, 'tube': 3})

        return IndirectSecondary(detectors, analyzers, a_per_d, sample_at, a, d)

    def triangulate_detectors(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate_detectors(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def triangulate_analyzers(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate_analyzers(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def mcstas_parameters(self, sample: Variable):
        from numpy import hstack
        from .combine import combine_parameters
        # pull out the list of 'distances', 'analyzer', 'detector', 'two_theta'
        # from each channel, and stack them into a single array per parameters
        parameters = combine_parameters(self.channels, sample)
        parameters['channel'] = hstack([channel.sample_space_angle(sample).value for channel in self.channels])
        return parameters

    def rtp_parameters(self, sample: Variable):
        from scipp import concat
        return [concat(q, dim='channel') for q in zip(*[c.rtp_parameters(sample) for c in self.channels])]

    def to_mccode(
            self,
            assembler: Assembler,
            sample: Instance,
            tank_bragg_monitor_group: str | None = None,
            tank_when: str | None = None,
            bragg_monitor_when: str | None =  None,
            settings: dict | None = None,
            **kwargs
    ):
        from scipp import vector, concat, max
        from ..mccode import ensure_user_var, ensure_registry
        ensure_registry(assembler, "mcdotstar/mcstas-slit-radial@main") # for slits
        ensure_user_var(assembler, 'int', 'secondary_cassette', 'Secondary spectrometer analyzer cassette index')

        origin = vector([0, 0, 0], unit='m')
        positions = [c.sample_space_angle(origin).to(unit='radian').value for c in self.channels]
        cov_xy = [c.coverage(origin, unit='radian') for c in self.channels]
        cov_x = 2 * max(concat([y for _, y in cov_xy], dim='channel')).value

        slits_name = 'slits'
        declared_positions = f'{slits_name}_positions'
        assembler.declare_array('double', declared_positions, positions, source=__file__, line=173)
        slits = assembler.component(slits_name, 'Slit_radial_multi', at=((0, 0, 0,), sample))
        slits.set_parameters(slit_width=cov_x, offset='slitAngle*DEG2RAD',
                             number=len(self.channels), radius='slitDistance', height=0.2,
                             positions=declared_positions)
        # `slit` is 0-8 iff scattered.
        # This could be `secondary_cassette = 1 + slit;` unambiguously
        slits.EXTEND("secondary_cassette = (SCATTERED) ? 1 + slit : -1;")

        # Insert the Bragg Peak elastic monitor
        mon = self.monitor.to_mccode(assembler, at=sample)

        if tank_bragg_monitor_group is not None:
            # One way to allow scattering from one sample to both elastic and inelastic
            # detectors.
            slits.GROUP(tank_bragg_monitor_group)
            mon.GROUP(tank_bragg_monitor_group)
        else:
            if bragg_monitor_when is not None:
                mon.WHEN(bragg_monitor_when)
            if tank_when is not None:
                slits.WHEN(tank_when)


        for index, channel in enumerate(self.channels):
            name = f"channel_{1 + index}"
            when = f"{1 + index} == secondary_cassette"
            channel.to_mccode(assembler, sample, name=name, when=when, settings=settings, **kwargs)



    def add_to_graph(self, upstream: str | None, name: str, graph: DiGraph):
        graph.add_node('slits')
        if upstream is not None:
            graph.add_edge(upstream, 'slits')
        cs = [channel.add_to_graph('slits', f"channel_{1 + index}", graph) for index, channel in enumerate(self.channels)]
        mn = self.monitor.add_to_graph(upstream, self.monitor.name, graph)
        return [*cs, mn]


from mccode_antlr.assembler import Assembler
from mccode_antlr.common import InstrumentParameter


def ensure_user_var(a: Assembler, dtype: str, name: str, description: str):
    """Ensure that particle struct parameter is declared"""
    a.ensure_user_var(f'{dtype} {name}; // {description}')


def declare_array(instrument: Assembler, element_type: str, name: str, description: str, values):
    instrument.declare_array(element_type, name, values)


def ensure_parameter(a: Assembler, data_type: str, name: str, description: str):
    """Ensure that a parameter is declared in the instrument"""
    a.parameter(f'{data_type} {name}; // {description}', ignore_repeated=True)


def ensure_runtime_line(a: Assembler, line: str):
    """Ensure that a runtime-defined parameter is declared in the instrument

        Utilizes the parser to define an `InstrumentParameter` object, which can have
        a string representation like:
            parameter_type parameter_name/"parameter unit" = default_value; // comment
        which gets truncated at the ';' (or '//' if ';' is missing) such that the comment
        is dropped.

        Each of parameter_type, parameter_unit, default_value and comment are optional,
        such that
            parameter_name
        is the minimally valid definition.
        If missing, `parameter_type` is equivalent to specifying 'double'.
        The default value depends on the type, but is typically 0.

        Repeated definitions of the same parameter are ignored, but inconsistent
        default, type, or units will raise an error.
    """
    return ensure_runtime_parameter(a, InstrumentParameter.parse(line))


def ensure_runtime_parameter(a: Assembler, par: InstrumentParameter):
    """Ensure that a runtime-defined parameter is declared in the instrument

    Repeated definitions of the same parameter are ignored, but inconsistent
    default, type, or units will raise an error.
    """
    held = a.instrument.get_parameter(par.name)
    if held := a.instrument.get_parameter(par.name) and held != par:
        msg = f"Parameter {par.name} already defined"
        if held.value is not None:
            msg += f" with value {held.value}"
        if held.unit is not None:
            msg += f" {held.unit}"
        raise RuntimeError(msg)
    else:
        a.instrument.add_parameter(par)


def ensure_registry(a: Assembler, specification: str):
    """Ensure that a register-defined parameter is declared in the instrument

    Parameters
    ----------
    a : Assembler
    specification : str
        Any of the mccode_antlr.reader.registry supported formats,
        1. ``{resolvable folder path}``
        2. ``{name} {resolvable folder path}``
        3. ``{name} {resolvable url} {resolvable file path}``
        4. ``{name} {resolvable url} {version} {registry file name}``
        5. ``git+{url}@{version}`` or ``git+{url}@{version}#{registry-file}``
        6. ``{owner}/{repo}@{version}`` or ``{owner}/{repo}@{version}#{registry-file}``
    """
    from mccode_antlr.reader.registry import registry_from_specification as rfs
    reg = rfs(specification)
    if not reg:
        msg = f"Unable to construct registry from {specification}"
        if not 'http' in specification and '/' in specification and not '@' in specification:
            msg += " missing @{tag|branch|commit} in GitHub Actions style specification"
        raise RuntimeError(msg)
    if not reg in a.instrument.registries:
        a.instrument.registries += (reg,)


# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT
from operator import getitem


def test_bifrost_whole():
    from niess.bifrost.parameters import primary_parameters, tank_parameters
    from niess.bifrost import Tank, Primary

    primary = Primary.from_calibration(primary_parameters())
    tank = Tank.from_calibration(tank_parameters())


def test_bifrost_mccode():
    from niess.bifrost.parameters import primary_parameters, tank_parameters
    from niess.bifrost import Tank, Primary
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler
    from pathlib import Path

    bifrost = Assembler('bifrost', flavor=Flavor.MCSTAS)

    primary = Primary.from_calibration(primary_parameters())
    primary.to_mccode(bifrost)

    # TODO insert pre- and post-sample things here
    #      e.g., the split_at location at the end of the guide
    #      any filters, e.g., a hits-the-sample MCPL filter, or a Be-transmission filter
    #      the radial collimator between sample and tank, etc.

    tank = Tank.from_calibration(tank_parameters())
    tank.to_mccode(bifrost, 'sample_origin')

    # The following test is extremely fragile and only useful when making possibly
    # breaking changes for local testing. The CI does not enforce the McCode version
    # use for fetched components, so this is very likely to break.
    #
    # from mccode_antlr.io.json import load_json
    # from msgspec.structs import fields
    #
    # # The loaded-from-state Instr should be equivalent across most niess changes
    # # but may break with changes in mccode-antlr.
    # # It is therefore up to you to decide if a new bifrost_assembler.json should be
    # # minted -- Good luck!
    # instr = load_json(Path(__file__).parent / "bifrost_assembler.json")
    #
    # # All members of the instrument should be the same, but registry equivalency
    # # is too strict since the order is not terribly important for a built-instr.
    # for par in fields(instr):
    #     if par.name != 'registries':
    #         assert getattr(instr, par.name) == getattr(bifrost.instrument, par.name)
    # #
    # # # If the above failed due to upstream changes, you probably want to stash any
    # # # modifications you've made, then come back here, disable the above and enable
    # # # the following:
    # # from mccode_antlr.io.json import save_json
    # # save_json(Path(__file__).parent / "bifrost_assembler.json", bifrost.instrument)
    # # # Then restore your changes and re-enable the checks above.


# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT


def test_bifrost_tank_importable():
    from importlib.util import find_spec
    if not find_spec('niess'):
        raise RuntimeError('No niess available!')
    if not find_spec('niess.bifrost'):
        raise RuntimeError('No niess.bifrost available!')
    from niess.bifrost import Tank


def test_bifrost_tank_constructable():
    from niess.bifrost import Tank
    tank = Tank.from_calibration()
    assert len(tank.channels) == 9


def test_bifrost_tank_calibratable():
    from scipp import scalar, isclose
    from niess.bifrost import Tank
    from niess.bifrost.parameters import tank_parameters
    calibration = tank_parameters()
    tank = Tank.from_calibration(calibration)

    assert len(tank.channels) == 9

    vertical = 4.0, 3.674467758121384, 3.372105326299701, 3.1338895319560107, 2.9399369139099463
    horizontal = scalar(5.2, unit='deg')

    params = calibration['channels']
    for channel in tank.channels:
        assert len(channel.pairs) == 5
        for index, (analyzer, triplet) in enumerate((arm.analyzer, arm.detector) for arm in channel.pairs):
            assert analyzer.count == params['blade_count']['analyzer', index]
            hor_cov, ver_cov = analyzer.coverage(params['sample'], unit='deg')

            # The coverage should approximately match that specified in the calibration
            # For 2.7 meV, the vertical coverage is +/- 2 degrees (e.g., 4 degrees)
            # Higher final energies are smaller to have constant Q-perpendicular
            # coverage, with exact values following `vertical` above ... or calculated
            # from the ratio of kf values
            expected = scalar(vertical[index], unit='deg').to(unit=ver_cov.unit)
            assert ver_cov >= expected
            assert isclose(ver_cov, expected, atol=scalar(0.1, unit='deg'))
            # The analyzers have sufficient width to fully illuminate their detector
            # tubes, which means they are _wider_ than the detectors to account for
            # the finite crystal mosaic.
            mosaic = analyzer.central_blade.mosaic.to(unit='deg')
            # The detectors each cover (at least) 5.2 degrees horizontally for all energies
            # So the analyzer should be at least 5.2 + mosaic
            expected = mosaic + horizontal.to(unit=hor_cov.unit)
            assert hor_cov >= expected
            assert isclose(hor_cov, expected, atol=scalar(1.2, unit='deg'))

            # We can also extract the detector effective horizontal coverage
            # since we know the sample-to-analyzer and analyzer-to-detector distances:
            arm_hor, arm_ver = channel.pairs[index].coverage(params['sample'], unit='deg')
            # these should actually be identical
            assert isclose(arm_ver, ver_cov)
            # and the horizontal coverage is calculated from the tube length,
            # which should be much closer to the expected value of 5.2 degrees
            # but since the current parameters are the total length instead of the
            # _active_ length, this coverage is slightly too big
            assert isclose(arm_hor, horizontal, atol=scalar(0.6, unit='deg'))


def test_bifrost_tank_angles():
    """The main purpose of this test is to show how to get the (horizontal)
    scattering limit boundaries for a single setting of the tank position"""
    from scipp import vector, concat, scalar
    from niess.bifrost import Tank
    from niess.bifrost.parameters import tank_parameters
    params = tank_parameters()
    assert 'channels' in params
    assert 'elastic_monitor' in params

    tank = Tank.from_calibration(params)

    sample = vector(value=[0, 0, 0], unit='m')

    def a4_limits(ch):
        center = ch.sample_space_angle(sample).to(unit='deg')
        da4, _ = ch.coverage(sample, unit='deg')
        return center - da4/2, center + da4/2

    limits = [a4_limits(ch) for ch in tank.channels]

    flat = concat([x for l in limits for x in l], 'limits')
    # verify that the limits increase monotonically
    changes = flat[1:] - flat[:-1]
    assert all(c > scalar(0, unit='deg') for c in changes)



# from dataclasses import dataclass, fields
from __future__ import annotations

import msgspec
from networkx import DiGraph
from msgspec.structs import fields
from functools import lru_cache
from ..utilities import calibration

# TODO Consider whether it would be possible to use any one of:
#      datclass-wizzard: https://dataclass-wizard.readthedocs.io
#      pydantic: https://docs.pydantic.dev
#      dacite: https://github.com/konradhalas/dacite
#      to handle (de)serializing these nested dataclass objects from calibration data.

# @dataclass
class Section(msgspec.Struct, tag=True):

    @classmethod
    @lru_cache(maxsize=None)
    def parts(cls):
        """Get the ordered list of components which make up this Section"""
        return [fi.name for fi in fields(cls)]

    @classmethod
    @lru_cache(maxsize=None)
    def types(cls):
        """Get the ordered list of component types which make up this Section"""
        return [fi.type for fi in fields(cls)]

    @classmethod
    @lru_cache(maxsize=None)
    def items(cls):
        """Get the ordered list of component names and types which make up this Section"""
        return [(n, t) for n, t in zip(cls.parts(), cls.types())]

    @classmethod
    @lru_cache(maxsize=None)
    def field_types(cls):
        return {fi.name: fi.type for fi in fields(cls)}

    def to_mccode(self, *args, **kwargs):
        for part in self.parts():
            getattr(self, part).to_mccode(*args, **kwargs)

    @classmethod
    @calibration
    def from_calibration(cls, parameters: dict):
        for part in cls.parts():
            assert part in parameters

        def named_par(name):
            if 'name' not in parameters[name]:
                parameters[name]['name'] = name
            return parameters[name]

        def to_type(name):
            typ = cls.field_types()[name]
            if isinstance(parameters[name], typ):
                return parameters[name]
            return typ.from_calibration(named_par(name))

        return cls(*[to_type(n) for n in cls.parts()])

    def to_dict(self):
        return {f: getattr(self, f) for f in self.parts()}

    @classmethod
    def from_dict(cls, d):
        # TODO make a function to recurse through a dictionary and ensure
        #      scipp.Variable and mccode_antlr.? are converted
        return cls.from_calibration(d)

    def to_graph(self):
        from networkx import DiGraph
        graph = DiGraph()
        self.add_to_graph(None, '', graph)
        return graph

    def add_to_graph(self, upstream: str | None, unused_section_name: str, graph: DiGraph):
        last = upstream
        for name in self.__struct_fields__:
            names = getattr(self, name).add_to_graph(last, name, graph)
            if isinstance(names, list) and len(names) == 1:
                last = names[0]
        return [last]


from mccode_antlr.assembler import Assembler
from mccode_antlr.instr.instance import Instance
from scipp import Variable
from .component import Component

# Monitors used on BIFROST, which are limited within McStas to always produce histograms

def nexus_structure_metadata(topic: str, source: str, *, variables, constants):
    from json import dumps
    from mccode_antlr.common import MetaData
    from mccode_to_kafka.writer import da00_dataarray_config
    # TODO update eniius.utils.mccode_component_eniius_data to use ('*', 'application/json') instead of
    #      only ('eniius_data', 'json')
    # The correct JSON to encode is a dictionary with a single key, 'data', which itself is a dictionary with
    # the keys 'type' and 'value'. The value of 'type' is 'dict' and the value of 'value' is the actual NeXus structure.
    # This is very hacky and is (currently) necessary to escape parsing within eniius.

    struct = da00_dataarray_config(topic, source=source, variables=variables, constants=constants)

    return MetaData.from_instance_tokens(topic, 'application/json', 'nexus_structure_stream_data', dumps(struct))


def add_monitor_metadata(instr_name: str, inst: Instance, n):
    from mccode_to_kafka.writer import da00_variable_config
    axes = {
        'signal': {'unit': 'counts', 'label': f'{inst.name} counts', 'shape': [n]},
        'errors': {'unit': 'counts', 'label': f'{inst.name} count errors',
                   'shape': [n]},
        't': {'unit': 'microsecond', 'label': 'time since reference', 'shape': [n + 1]},
    }
    configs = {
        k: da00_variable_config(**v, name=k, axes=['t'], data_type='float64')
        for k, v in axes.items()
    }
    variables = [configs['signal'], configs['errors']]
    constants = [configs['t']]
    topic = f'{instr_name.lower()}_beam_monitor'
    source = inst.name  # 'cbm1', 'cbm2', etc. in the real instrument, for now
    metadata = nexus_structure_metadata(
        topic=topic, source=source, variables=variables, constants=constants
    )
    inst.add_metadata(metadata)
    return inst


class FrameMonitor(Component):
    @staticmethod
    def time_bins():
        return int(1e6 / 14.0 / 7) # 7 microsecond bins

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        from niess.mccode import ensure_registry
        ensure_registry(assembler, 'mcdotstar/mcstas-frame-tof-monitor@main')

        inst = super().to_mccode(assembler, at, rotate)
        # Build the NeXus Structure entry to point to the correct Kafka stream
        return add_monitor_metadata(assembler.name, inst, self.time_bins())

    def __partial__mccode__(self) -> tuple[str, dict]:
        return 'Frame_monitor', {'nt': self.time_bins(), 'frequency': 14.0}


class FissionChamber(FrameMonitor):
    """Zero-dimensional fission chamber monitor.
    Outputs events without any spatial information.
    """
    width: Variable
    height: Variable
    thickness: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        return cls(
            name=name,
            position=position,
            orientation=orientation,
            width=width,
            height=height,
            thickness=thickness
        )

    def __mccode__(self) -> tuple[str, dict]:
        t, p = self.__partial__mccode__()
        p.update({
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
        })
        return t, p


class He3Monitor(FrameMonitor):
    """Zero-dimensional He3 tube monitor.
    Outputs events without any spatial information.
    """
    radius: Variable
    length: Variable
    pressure: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        radius = cal['radius']
        length = cal['length']
        pressure = cal['pressure']
        return cls(
            name=name, position=position, orientation=orientation,
            radius=radius, length=length, pressure=pressure
        )

    def __mccode__(self) -> tuple[str, dict]:
        t, p = self.__partial__mccode__()
        p.update({
            'xwidth': 2 * self.radius.to(unit='m').value,
            'yheight': self.length.to(unit='m').value,
        })
        return t, p


class BeamCurrentMonitor(FrameMonitor):
    """Zero-dimensional beam current monitor.
    Outputs a current sampled at a configurable frequency.
    """
    width: Variable
    height: Variable
    thickness: Variable
    sample_rate: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        sample_rate = cal.get('sample_rate', cal.get('frequency'))
        if sample_rate is None:
            raise ValueError(f'The sample rate for {name} must be defined')
        return cls(
            name=name, position=position, orientation=orientation,
            width=width, height=height, thickness=thickness, sample_rate=sample_rate
        )

    def time_bins(self):
        from scipp import scalar
        source_frequency = scalar(14.0, unit='Hz')
        return int((self.sample_rate.to(unit='Hz') / source_frequency).value)

    def __mccode__(self) -> tuple[str, dict]:
        t, p = self.__partial__mccode__()
        p.update({
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
        })
        return t, p


class GEM2D(FrameMonitor):
    """Two-dimensional Gas Electron Multiplier monitor.
    Outputs events on X or Y strips (without coincidence) or at (X, Y) point
    (with coincidence).

    For now the pixelation is ignored and a 0-D no-coincidence frame monitor is output
    """
    width: Variable
    height: Variable
    thickness: Variable
    x_strips: int
    y_strips: int

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        x_strips = cal.get('x_strips', cal.get('nx', 1))
        y_strips = cal.get('y_strips', cal.get('ny', 1))
        return cls(
            name=name, position=position, orientation=orientation,
            width=width, height=height, thickness=thickness,
            x_strips=x_strips, y_strips=y_strips
        )

    def __mccode__(self) -> tuple[str, dict]:
        t, p = self.__partial__mccode__()
        p.update({
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
        })
        return t, p



# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT
from .component import Base
from scipp import Variable


class IdealCrystal(Base):
    position: Variable
    tau: Variable

    def triangulate(self, unit=None):
        from scipp import sqrt, dot, vector, arange, concat, cross, isclose
        from scipp.spatial import rotations_from_rotvecs
        if unit is None:
            unit = self.position.unit
        lt = sqrt(dot(self.tau, self.tau))
        # *a* vector perpendicular to tau
        p = cross(self.tau, vector([1., 0, 0]) if isclose(self.tau.fields.z, lt) else vector([0, 0, 1.]))
        p = (p/sqrt(dot(p, p)) / lt).to(unit=unit)
        a = arange(start=0, stop=360, step=10, dim='vertices', unit='degree')
        r = rotations_from_rotvecs(a*self.tau/lt)
        vertices = concat((self.position, r*p + self.position), dim='vertices')
        lv = len(r)
        triangles = [[0, i + 1, (i + 1)%lv + 1] for i in range(lv)]

        return vertices, triangles

    def extreme_path_corners(self, horizontal: Variable, vertical: Variable, unit=None):
        if unit is None:
            unit = self.position.unit
        return self.position.to(unit=unit)

    def __eq__(self, other):
        if not isinstance(other, IdealCrystal):
            return False
        return self.position == other.position and self.tau == other.tau

    def approx(self, other):
        from scipp import allclose
        if not isinstance(other, IdealCrystal):
            return False
        return allclose(self.position, other.position) and allclose(self.tau, other.tau)
    
    def __post_init__(self):
        from scipp import DType
        if self.position.dtype != DType.vector3:
            raise RuntimeError("position must be of type scipp.DType('vector3')")
        if self.tau.dtype != DType.vector3:
            raise RuntimeError("tau must be of type scipp.DType('vector3')")

    @property
    def momentum(self) -> Variable:
        from scipp import norm
        return norm(self.tau)

    @property
    def momentum_vector(self) -> Variable:
        return -self.tau

    @property
    def plane_spacing(self) -> Variable:
        from math import pi
        return 2 * pi / self.momentum

    def scattering_angle(self, **kwargs) -> Variable:
        from math import pi, inf
        from scipp import asin, scalar, isinf, isnan, abs, norm
        from ..spatial import __is_vector__
        if len(kwargs) != 1:
            raise RuntimeError("A single keyword argument (k, wavenumber, wavelength) is required")
        k = kwargs.get('k', kwargs.get('wavenumber', 2 * pi / kwargs.get('wavelength', scalar(inf, unit='angstrom'))))
        if __is_vector__(k):
            k = norm(k)
        if k.value == 0 or isinf(k) or isnan(k):
            raise RuntimeError("The provided keyword must produce a finite wavenumber")
        t = self.momentum.to(unit=k.unit)
        if t > 2 * abs(k):
            raise RuntimeError(f"Bragg scattering from |Q|={t:c} planes is not possible for k={k:c}")
        return 2 * asin(t / (2 * k))

    def wavenumber(self, scattering_angle: Variable):
        from scipp import sin
        return self.momentum / (2 * sin(0.5 * scattering_angle.to(unit='radian')))

    def reflectivity(self, *a, **k) -> float:
        return 1.

    def transmission(self, *a, **k) -> float:
        return 1.

    def rtp_parameters(self, sample: Variable, center: Variable, out_of_plane: Variable):
        from scipp import cross, dot, sqrt, atan2
        y = cross(out_of_plane, center - sample)
        y /= sqrt(dot(y, y))
        x = cross(y, out_of_plane)
        x /= sqrt(dot(x, x))

        pc = self.position - center
        rtp_x = dot(pc, x)
        rtp_y = dot(pc, y)
        rtp_angle = atan2(x=dot(self.tau, -y), y=dot(self.tau, x))
        return rtp_x, rtp_y, rtp_angle


class Crystal(IdealCrystal):
    shape: Variable  # lengths: (in-scattering-plane perpendicular to Q, perpendicular to plane, along Q)
    orientation: Variable
    mosaic: Variable

    def triangulate(self, unit=None):
        from ..spatial import vector_to_vector_quaternion
        from scipp import vectors, vector
        if unit is None:
            unit = self.position.unit
        r = vector_to_vector_quaternion(vector([0, 0, 1.]), self.tau)
        x, y, z = 0.5 * self.shape.value
        vertices = vectors(unit=self.shape.unit, dims=['vertices'],
                          values=[[-x, -y, -z], [+x, -y, -z], [+x, +y, -z], [-x, +y, -z],
                                  [-x, -y, +z], [+x, -y, +z], [+x, +y, +z], [-x, +y, +z]])
        vertices = r * self.orientation * vertices
        faces = [[0, 2, 1], [2, 0, 3], [1, 2, 6], [1, 6, 5], [0, 1, 5], [0, 5, 4],
                 [3, 0, 4], [3, 4, 7], [2, 3, 7], [2, 7, 6], [4, 5, 6], [4, 6, 7]]
        return vertices.to(unit=unit) + self.position.to(unit=unit), faces

    def bounding_box(self, basis: Variable, unit=None):
        from ..spatial import bounding_box
        v, _ = self.triangulate(unit=unit)
        return bounding_box(v, basis)

    def __eq__(self, other):
        if not isinstance(other, Crystal):
            return False
        return self.shape == other.shape and super().__eq__(other)

    def approx(self, other):
        from scipp import allclose
        if not isinstance(other, Crystal):
            return False
        return allclose(self.shape, other.shape) and super().approx(other)

    def __post_init__(self):
        from scipp import DType
        super().__post_init__()
        if self.shape.dtype != DType.vector3:
            raise RuntimeError("shape must be of type scipp.DType('vector3')")
        if self.orientation.dtype != DType.rotation3:
            raise RuntimeError("orientation must be of type scipp.DType('rotation3')")

    def mcstas_parameters(self):
        from numpy import hstack
        return hstack(self.position.value, self.shape.value)



from __future__ import annotations

import msgspec
from msgspec.structs import fields
from functools import lru_cache
from scipp import Variable
from networkx import DiGraph
from mccode_antlr.assembler import Assembler
from mccode_antlr.instr import Instance

# TODO: Add a 'tag' property to either Base or Component
#       This is intended to represent an ESS Facility Breakdown Structure (FBS) tag,
#       and should take precedence over 'name' for the purpose of making component graphs

class Base(msgspec.Struct):

    @classmethod
    @lru_cache(maxsize=None)
    def fields(cls):
        return [fi.name for fi in fields(cls)]

    @classmethod
    @lru_cache(maxsize=None)
    def types(cls):
        """Get the ordered list of component types which make up this Section"""
        return [fi.type for fi in fields(cls)]

    @classmethod
    @lru_cache(maxsize=None)
    def items(cls):
        """Get the ordered list of component names and types which make up this Section"""
        return [(n, t) for n, t in zip(cls.fields(), cls.types())]

    def to_dict(self):
        return {k: getattr(self, k) for k in self.fields()}

    @classmethod
    def from_dict(cls, data):
        for k, t in cls.items():
            if k not in data:
                raise KeyError(f"{k} not found in data")
            if not isinstance(data[k], t) and isinstance(data[k], dict) and hasattr(t, 'from_dict'):
                data[k] = t.from_dict(data[k])
        return cls(**data)

    def __eq__(self, other):
        from scipp import identical
        if not isinstance(other, type(self)):
            return False
        for field in self.fields():
            a = getattr(self, field)
            b = getattr(other, field)
            if a is None or b is None:
                if a is not None or b is not None:
                    return False
            elif isinstance(a, Variable):
                if not identical(a, b, equal_nan=True):
                    return False
            else:
                if a != b:
                    return False
        return True

    def add_to_graph(self, upstream: str | None, name: str, graph: DiGraph):
        graph.add_node(name)
        if upstream is not None:
            graph.add_edge(upstream, name)
        return [name]


class Component(Base, kw_only=True):
    """Any component in the instrument.

    Note
    ----
    If an inheriting class adds an 'offset' attribute to the component, the
    position reported for McStas/McXtrace/McCode has that offset added to position

    Parameters
    ----------
    name: str
        The (unique) name of the component instance
    position: Vector
        The position of the component instance in a global coordinate system. This
        may differ from the position required for, e.g., McStas (see 'offset' Note).
    orientation: Quaternion
        The orientation of the component instance in scipp quaternion form. This
        transforms the coordinate system of the component into the global coordinate
        system.
    tag: str | None
        The Facility Breakdown Structure (FBS) tag representing this component
    """
    name: str
    position: Variable
    orientation: Variable
    # tag: str | None = None

    @classmethod
    def from_calibration(cls, calibration: dict):
        name = calibration['name']
        position = calibration['position']
        orientation = calibration['orientation']
        # tag = calibration.get('tag')
        # return cls(name=name, position=position, orientation=orientation, tag=tag)
        return cls(name=name, position=position, orientation=orientation)

    @classmethod
    def from_dict(cls, dictionary):
        return cls.from_calibration(dictionary)

    def __mccode__(self) -> tuple[str, dict]:
        """Return the component type name and parameters needed to produce a McCode instance"""
        return 'Arm', {}

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        from mccode_antlr.common.parameters import InstrumentParameter as InstPar
        from ..spatial import mccode_ordered_angles
        from ..mccode import ensure_runtime_parameter

        comp, pars = self.__mccode__()

        if len(pairs:=[(k, x) for k, x in pars.items() if isinstance(x, InstPar)]):
            for name, value in pairs:
                ensure_runtime_parameter(assembler, value)
                pars[name] = str(value)

        at_rel = 'ABSOLUTE' if at is None else at
        rot_rel = 'ABSOLUTE' if rotate is None else rotate

        at = self.position
        if hasattr(self, 'offset'):
            at += getattr(self, 'offset')
        at = (at.to(unit='m').value, at_rel)
        rot = (mccode_ordered_angles(self.orientation), rot_rel)

        return assembler.component(self.name, comp, at=at, rotate=rot, parameters=pars)



from scipp import Variable
from .component import Component
from mccode_antlr.instr import Instance
from mccode_antlr.assembler import Assembler


class Chopper(Component):
    """Any device which periodically opens a path that particles may traverse"""
    velocity: Variable  # angular velocity scalar or vector
    phase: Variable
    radius: Variable
    windows: Variable
    width: Variable  # the path width
    height: Variable  # the path height


class DiscChopper(Chopper):
    """Ideally infinitely thin material with rotation vector parallel to the path"""
    # `radius` is the outer dimension of the disc.
    # `windows` are ordered angular edges of the openings in the disc
    # `phase` is the time=0 angular orientation of the disc
    offset: Variable # the position of the path relative to the disc center

    @property
    def speed(self):
        from scipp import dot, vector
        from ..spatial import __is_vector__
        if not __is_vector__(self.velocity):
            return self.velocity.to(unit='Hz')
        # TODO verify the sense of this wrt the McStas definition
        return dot(vector(value=[0, 0, 1.]), self.velocity).to(unit='Hz')

    def chopper_lib_parameters(self):
        """Useful for specifying elements of a vector used by chopper-lib"""
        from scipp import max, min, norm
        speed_name = f'{self.name}speed'
        phase_name = f'{self.name}phase'
        if self.windows.size != 2:
            raise ValueError("chopper-lib expects only one window")
        angle = (max(self.windows) - min(self.windows)).to(unit='deg').value
        distance = norm(self.position).to(unit='m').value
        return '{' + f'{speed_name}, {phase_name}, {angle}, {distance}' + '}'

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import scalar, array, vector
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        velocity = cal.get('velocity', cal.get('frequency'))
        if velocity is None:
            raise ValueError('velocity (or frequency) cannot be None')
        phase = cal.get('phase', scalar(0.0, unit='deg'))
        radius = cal['radius']
        if 'windows' not in cal:
            cal['windows'] = cal['angle'].to(unit='deg') / 2 * array(values=[-1, 1], dims=['edges'])
        windows = cal['windows']
        width = cal.get('width') # None is actually acceptable
        height = cal.get('height')  # None is acceptable, then slit extends to center
        offset = cal.get('offset', vector([0, 0, 0], unit='m'))
        return cls(
            name=name,
            position=position,
            orientation=orientation,
            velocity=velocity,
            phase=phase,
            radius=radius,
            windows=windows,
            width=width,
            height=height,
            offset=offset
        )


    def __mccode__(self) -> tuple[str, dict]:
        from scipp import max, min
        if self.windows.size != 2:
            # The McStas way of handling multiple windows is one of two options:
            #   1. if they are equally spaced and all the same size, use 'nslit'
            #   2. otherwise use a group of choppers all at the same position but
            #      rotated relative to one another by their slit difference. One per.
            raise ValueError("Currently only one window supported. Investigate using a group here")
        params = {
            'theta_0': (max(self.windows) - min(self.windows)).to(unit='deg').value,
            'nslit': 1,
            'radius': self.radius.to(unit='m').value,
            'nu': f'{self.name}speed',
            'phase': f'{self.name}phase',
        }
        # Only add width of height if provided:
        if self.width is not None:
            params['xwidth'] = self.width.to(unit='m').value
        if self.height is not None:
            params['yheight'] = self.height.to(unit='m').value
        return 'DiskChopper', params

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        from ..mccode import ensure_runtime_line as ensure
        ensure(assembler, f'{self.name}speed/"Hz" = {self.speed.value}')
        ensure(assembler, f'{self.name}phase/"degree" = {self.phase.to(unit="deg").value}')
        # the offset is handled by super's to_mccode -- no problems.
        return super().to_mccode(assembler, at, rotate)


class FermiChopper(Chopper):
    """Ideally infinitely tall cylinder with a group of curved channels through
    its center, rotation vector parallel to its axis, and perpendicular to the path"""
    # `radius` describes the cylinder dimension
    # `windows` are the edges of the channels, as offsets from the central line
    # `phase` is the t=0 angular orientation of the central line wrt the path
    curvature: Variable


from scipp import Variable


def known_channel_params():
    from scipp import array, scalar, vector, vectors
    known = dict()
    dist_sa = {
        's': [1.100, 1.238, 1.342, 1.443, 1.557],
        'm': [1.189, 1.316, 1.420, 1.521, 1.623],
        'l': [1.276, 1.392, 1.497, 1.599, 1.701],
    }
    known['sample_analyzer_distance'] = {k: array(values=v, unit='m', dims=['analyzer']) for k, v in dist_sa.items()}
    known['analyzer_detector_distance'] = known['sample_analyzer_distance']['m']
    d_length_mm = {
        's': [217.9, 242.0, 260.8, 279.2, 298.8],
        'm': [226.0, 249.0, 267.9, 286.3, 304.8],
        'l': [233.9, 255.9, 274.9, 293.4, 311.9],
    }
    dex = scalar(10, unit='mm')  # The detector tubes were ordered with 10 mm extra length buffer
    known['detector_length'] = {k: dex + array(values=v, unit='mm', dims=['analyzer']) for k, v in d_length_mm.items()}
    known['detector_offset'] = vectors(values=[[0, 0, -14.], [0, 0, 0], [0, 0, 14]], unit='mm', dims=['tube'])
    known['detector_orient'] = vector([0, 0, 0], unit='mm')
    a_shape_mm = {
        's': [[12.0, 134, 1], [14.0, 147, 1], [11.5, 156, 1], [12.0, 165, 1], [13.5, 177, 1]],
        'm': [[12.5, 144, 1], [14.5, 156, 1], [11.5, 165, 1], [12.5, 174, 1], [13.5, 183, 1]],
        'l': [[13.5, 150, 1], [15.0, 162, 1], [12.0, 171, 1], [13.0, 180, 1], [14.0, 189, 1]],
    }
    known['crystal_shape'] = {k: vectors(values=v, unit='mm', dims=['analyzer']) for k, v in a_shape_mm.items()}
    known['crystal_mosaic'] = scalar(40., unit='arcminutes')
    known['blade_count'] = array(values=[7, 7, 9, 9, 9], dims=['analyzer'])  # two lowest energy analyzer have 7 blades
    known['d_spacing'] = scalar(3.355, unit='angstrom')  # PG(002)
    known['coverage'] = scalar(2., unit='degree') # +/- 2 degrees at 2.7 meV, constant delta-Q at higher energies
    known['energy'] = array(values=[2.7, 3.2, 3.8, 4.4, 5.], unit='meV', dims=['analyzer'])
    known['sample'] = vector([0, 0, 0.], unit='m')
    known['gap'] = array(values=[2, 2, 2, 2, 2.], unit='mm', dims=['analyzer'])
    known['variant'] = 'm'

    known['resistance'] = scalar(380., unit='Ohm')
    known['contact_resistance'] = scalar(88.0/2, unit='Ohm')
    known['resistivity'] = scalar(185., unit='Ohm/in').to(unit='Ohm/m')

    known['beryllium_filter_ncrystal_cfg'] = 'Be_sg194'
    known['beryllium_filter_temperature'] = scalar(77., unit='K')
    known['beryllium_filter_inner_radius'] = scalar(510., unit='mm')
    known['beryllium_filter_outer_radius'] = scalar(720., unit='mm')
    known['beryllium_filter_height'] = scalar(57.0, unit='mm')
    known['beryllium_filter_width'] = scalar(7.8, unit='deg')
    known['radial_collimator_inner_radius'] = scalar(500., unit='mm')
    known['radial_collimator_outer_radius'] = scalar(820., unit='mm')
    known['radial_collimator_height'] = scalar(65., unit='mm')
    known['radial_collimator_collimation'] = scalar(0.65, unit='deg')
    known['radial_collimator_width'] = scalar(7.8, unit='deg')

    return known

def tank_parameters():
    from scipp import scalar
    known = dict()
    known['channels'] = known_channel_params()
    known['sample_elastic_monitor_distance'] = scalar(800., unit='mm')
    known['tank_elastic_monitor_angle'] = scalar(59., unit='deg')
    known['elastic_monitor'] = {
        'name': 'elastic_monitor',
        'radius': scalar(1., unit='inch').to(unit='mm'),
        'length': scalar(3.2, unit='inch').to(unit='mm'),
        'pressure': scalar(0.2, unit='atm'),
    }
    return known


def tube_xz_displacement_to_quaternion(length: Variable, displacement: Variable):
    from scipp import vector, scalar, any, sqrt, allclose
    from ..spatial import vector_to_vector_quaternion
    com_to_end = length * vector([0, 0.5, 0]) + displacement
    l2 = length * length
    x2 = displacement.fields.x * displacement.fields.x
    z2 = displacement.fields.z * displacement.fields.z

    com_to_end.fields.y = sqrt(0.25 * l2 - x2 - z2)

    y2 = displacement.fields.y * displacement.fields.y
    if any(y2 > scalar(0, unit=y2.unit)) and not allclose(com_to_end.fields.y, 0.5 * length - displacement.fields.y):
        raise RuntimeError("Provided tube-end displacement vector(s) contain wrong y-component value(s)")

    # The tube *should* point along y, but we were told it is displaced in x and z;
    # return the orienting Quaternion that takes (010) to the actual orientation
    quaternion = vector_to_vector_quaternion(vector([0, 1, 0]), com_to_end)
    return quaternion


def primary_parameters(use_tcs=False):
    from scipp import array, vector, scalar, norm
    from scipp.spatial import rotations_from_rotvecs as r
    from ..spatial import mccode_quaternion, at_relative_dict, at_relative
    from .guide_compressor import primary_compressor_parameters
    from .guide_curved import curved_guide_parameters
    from .guide_expanding import expanding_guide_parameters
    from .guide_straight import straight_guide_parameters
    from .guide_closing import closing_guide_parameters
    p = dict()

    m = scalar(1.0, unit='m')
    mm = scalar(1.0, unit='mm').to(unit='m')
    z = vector([0, 0, 1.0])

    eps = 1.0e-5
    # The guide is defined in the 'w4' coordinate system
    guide_zero = vector([0.01277, 0, 1.903398 - eps], unit='m')
    guide_zero_rot = mccode_quaternion(0, -0.56, 0)

    if use_tcs:
        # Relative to the Target Coordinate System (TCS),
        # The BIFROST Instrument Specific Coordinate System (ISCS) is at:
        tcs_iscs_position = vector([-77.05, 67.72, 137], unit='mm').to(unit='m')
        tcs_iscs_orientation = r(vector([0, 132.14, 0], unit='deg'))
        # But the McStas ESS_butterfly component 'sector': 'W', 'beamline': 4
        # defines its own coordinate system, which is rotated relative to both
        # The ISCS position _in_ the McStas 'W4' coordinates:
        w4_iscs_position = vector([31.38,0,-0.01], unit='mm')
        # In thE TCS coordinate system, W4 focal point is at and oriented:
        w4_orientation = r(vector([0., 132.7, 0.],  unit='deg'))
        w4_position = vector([89, 137., -54], unit='mm').to(unit='m')
        # but we want to move everything to TCS:
        moderator_offset =vector([0, 137., 0], unit='mm').to(unit='m')
        guide_zero_rot = w4_orientation * guide_zero_rot
        guide_zero = at_relative(w4_position, w4_orientation, guide_zero)
    else:
        w4_position = vector([0, 0, 0.], unit='m')
        w4_orientation = r(vector([0, 0, 0.], unit='deg'))

    p['source'] = {
        'sector': 'W',
        'beamline': 4,
        'wavelength_minimum': 'source_lambda_min/"angstrom" = 0.75',
        'wavelength_maximum': 'source_lambda_max/"angstrom" = 30.0',
        'focus_distance': norm(guide_zero),
        'focus_width': (0.068797 + 2 * 0.01277) * m,  # including the substrate?
        'focus_height': 0.03472 * m,
        'position': w4_position,
        'orientation': w4_orientation,
    }

    p.update(primary_compressor_parameters(guide_zero, guide_zero_rot))

    bunker_chopper_height = scalar(0.047514 + 2 * 0.00331, unit='m')  # 2 * margin of error for floor settling in bunker
    hall_chopper_height = scalar(0.09 + 2 * 0.00423, unit='m')  # 2 * margin of error for piles settling under the long guide hall
    radius = 350 * mm
    offset = -(radius - bunker_chopper_height / 2) * vector([0, 1., 0])
    p['pulse_shaping_chopper_1'] = {
        'position': at_relative(p['nose']['end'], p['nose']['orientation'], (0.0306 * m) * z) - offset,
        'orientation': p['nose']['orientation'],
        'radius': radius,
        'height': bunker_chopper_height,
        'angle': scalar(170., unit='deg'),
        'frequency': scalar(14., unit='Hz'),
        'phase': scalar(0., unit='deg'),
        'offset': offset,
    }
    p['pulse_shaping_chopper_2'] = {
        'position': at_relative_dict(p['pulse_shaping_chopper_1'], (0.049 * m) * z),
        'orientation': p['pulse_shaping_chopper_1']['orientation'],
        'radius': radius,
        'height': bunker_chopper_height,
        'angle': scalar(170., unit='deg'),
        'frequency': scalar(14., unit='Hz'),
        'phase': scalar(0., unit='deg'),
        'offset': offset,
    }

    # From BIFROST Table of Optics, ESS-4813238 the end of subsystem 1 to the start
    # of subsystem 2 is exactly 81.707 mm. Private communication indicates instead
    # the gap from the end of the nose-guide to the start of the first curved guide
    # segment inside the PSC housing is 84 mm
    element_6_to_element_5 = 84 * mm
    rel_r = p['nose']['orientation']
    rel_p = at_relative(p['nose']['end'], rel_r, element_6_to_element_5 * z)

    # Element 6 in the old McStas instrument is the curved section. It includes
    # the first and second frame overlap choppers and a copper-substrate 'collimation'
    # section which helps prevent streaming
    el6, rel_p, rel_r = curved_guide_parameters(rel_p, rel_r, bunker_chopper_height)
    p.update(el6)

    # Following the curved section is the expanding 'connector' section including
    # the bunker wall feed through. There is a monitor but no choppers here
    ex, rel_p, rel_r = expanding_guide_parameters(rel_p, rel_r)
    p.update(ex)

    # The straight section includes the bandwidth choppers and a monitor
    ex, rel_p, rel_r = straight_guide_parameters(rel_p, rel_r, hall_chopper_height)
    p.update(ex)

    # The closing section focuses the beam and includes divergence limiting jaws
    ex, rel_p, rel_r = closing_guide_parameters(rel_p, rel_r)
    p.update(ex)
    # The reference point is now the downstream side of the final window.
    # The sample is exactly 578 mm from this position, but there are other things before

    # directly after the guide is an exchangeable B4C mask-aperture.
    p['mask'] = {
        'position': at_relative(rel_p, rel_r, (5 * mm) * z),
        'orientation': rel_r,
        'width': 50 * mm,
        'height': 60 * mm,
    }
    # then the normalization monitor
    p['normalization_monitor'] = {
        'position': at_relative(rel_p, rel_r, (15 * mm) * z),
        'orientation': rel_r,
        'width': 70 * mm,
        'height': 70 * mm,
        'thickness': 0.1 * mm,
    }
    # and finally a driven sample slit (that also can be adjusted along the beam)
    p['slit'] = {
        'position': at_relative(rel_p, rel_r, (30 * mm) * z),
        'orientation': rel_r,
        'width': 70 * mm,
        'height': 70 * mm,
    }

    # The primary spectrometer is followed by the sample stack, then secondary spectrometer
    # the McStas instrument will need numerous extra components around the sample

    p['sample_origin'] = {
        'position': at_relative(rel_p, rel_r, (578 * mm) * z),
        'orientation': rel_r,
    }

    return p


from __future__ import annotations

from typing import Union
from mccode_antlr.assembler import Assembler
from mccode_antlr.instr import Instance
from networkx import DiGraph
from scipp import Variable
from .component import Base, Component


def float_or_tuple_float(x):
    if isinstance(x, float):
        return x
    if isinstance(x, list) and all(isinstance(i, float) for i in x):
        return tuple(x)
    if isinstance(x, tuple) and all(isinstance(i, float) for i in x):
        return x
    raise ValueError("Only float values (or a list-like collection of floats) allowed")


def list_or_dict_to_type_list(typ, value):
    ret = []
    if isinstance(value, dict):
        for name, cdict in value.items():
            if not isinstance(cdict, dict):
                raise ValueError("Only dicts should be provided here")
            cdict['name'] = cdict.get('name', name)
            ret.append(typ.from_calibration(cdict))
    elif isinstance(value, list):
        for v in value:
            if not isinstance(v, dict):
                raise ValueError("Only dicts should be provided here")
            ret.append(typ.from_calibration(v))
    else:
        raise ValueError("Only dict or list of dicts should be provided here")
    return ret


class Guide(Component):
    length: Variable
    left: Union[float, tuple[float]]  # m-value for x > 0 face
    right: Union[float, tuple[float]] # m-value for x < 0 face
    top: Union[float, tuple[float]]  # m-value for y > 0 face
    bottom: Union[float, tuple[float]]   # m-value for the y < 0 face


class StraightGuide(Guide):
    width: Variable
    height: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='deg')))
        length = cal['length']
        m = cal.get('m', 1.0)
        left = cal.get('left', m)
        right = cal.get('right', m)
        top = cal.get('top', m)
        bottom = cal.get('bottom', m)
        if any(isinstance(x, (tuple, list)) for x in (left, right, top, bottom)):
            raise ValueError('StraightGuide does not support multiple m values')
        width = cal['width']
        height = cal['height']
        return cls(
            name=name,
            position=position,
            orientation=orientation,
            length=length,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
            width=width,
            height=height
        )

    def __mccode__(self) -> tuple[str, dict]:
        p = {
            'w1': self.width.to(unit='m').value,
            'h1': self.height.to(unit='m').value,
            'l': self.length.to(unit='m').value,
            'mleft': self.left,
            'mright': self.right,
            'mtop': self.top,
            'mbottom': self.bottom,
            'R0': 0.99,
            'Qc': 0.0217,
            'alpha': 3.1,
            'W': 0.003,
            'G': -9.82
        }
        p['w2'] = p['w1']
        p['h2'] = p['h1']
        return 'Guide_gravity', p

class SegmentedGuide(Base):
    name: str
    segments: list

    def add_to_graph(self, upstream: str | None, name: str, graph: DiGraph):
        last = upstream
        for x in self.segments:
            names = x.add_to_graph(last, x.name, graph)
            if len(names) == 1:
                last = names[0]
        return [last]

class StraightGuides(SegmentedGuide):
    segments: list[StraightGuide]

    @classmethod
    def from_calibration(cls, cal: dict):
        """Convert a dictionary of dictionaries to a list of StraightGuides

        This is likely called from Section.from_calibration, which would insert
        the *Section* name as an entry in the dictionary. This is the list's parent
        which we can store separately.
        """
        parent = cal.pop('name')
        segments = []
        if 'segments' in cal and len(cal) == 1:
            segments = list_or_dict_to_type_list(StraightGuide, cal['segments'])
        else:
            segments = list_or_dict_to_type_list(StraightGuide, cal)
        # for name, cdict in cal.items():
        #     if not isinstance(cdict, dict):
        #         raise ValueError("Only dicts should be in this calibration dictionary")
        #     if 'name' not in cdict:
        #         cdict['name'] = name
        #     segments.append(StraightGuide.from_calibration(cdict))
        return cls(parent, segments)

    def to_mccode(self, assembler: Assembler):
        for segment in self.segments:
            segment.to_mccode(assembler)


class TaperedGuide(Guide):
    in_width: Variable
    in_height: Variable
    out_width: Variable
    out_height: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='deg')))
        length = cal['length']
        m = cal.get('m', 1.0)
        left = cal.get('left', m)
        right = cal.get('right', m)
        top = cal.get('top', m)
        bottom = cal.get('bottom', m)
        if any(isinstance(x, (tuple, list)) for x in (left, right, top, bottom)):
            raise ValueError('StraightGuide does not support tuple m values')
        in_width = cal.get('in_width', cal.get('width'))
        out_width = cal.get('out_width', cal.get('width'))
        in_height = cal.get('in_height', cal.get('height'))
        out_height = cal.get('out_height', cal.get('height'))
        return cls(
            name=name,
            position=position,
            orientation=orientation,
            length=length,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
            in_width=in_width,
            out_width=out_width,
            in_height=in_height,
            out_height=out_height
        )

    def __mccode__(self) -> tuple[str, dict]:
        p = {
            'w1': self.in_width.to(unit='m').value,
            'w2': self.out_width.to(unit='m').value,
            'h1': self.in_height.to(unit='m').value,
            'h2': self.out_height.to(unit='m').value,
            'l': self.length.to(unit='m').value,
            'mleft': self.left,
            'mright': self.right,
            'mtop': self.top,
            'mbottom': self.bottom,
            'R0': 0.99,
            'Qc': 0.0217,
            'alpha': 3.1,
            'W': 0.003,
            'G': -9.82
        }
        return 'Guide_gravity', p


class TaperedGuides(SegmentedGuide):
    segments: list[TaperedGuide]

    @classmethod
    def from_calibration(cls, cal: dict):
        parent = cal.pop('name')
        segments = []
        if 'segments' in cal and len(cal) == 1:
            cal = cal['segments']
        for name, cdict in cal.items():
            if 'name' not in cdict:
                cdict['name'] = name
            segments.append(TaperedGuide.from_calibration(cdict))
        return cls(parent, segments)

    def to_mccode(self, assembler: Assembler):
        for segment in self.segments:
            segment.to_mccode(assembler)


class PartialEllipse(Base):
    major: Variable
    minor: Variable
    offset: Variable

    @classmethod
    def from_calibration(cls, length: Variable, cal: dict):
        from scipp import sqrt
        if not all(x in cal for x in ('major', 'minor', 'offset')):
            # calculate the ellipse parameters from the provided information
            all_of = 'in', 'out'
            any_of = 'entry', 'exit', 'midpoint'
            if not (all(x in cal for x in all_of) and any(x in cal for x in any_of)):
                raise ValueError()
            i = cal['in']  # distance from focal point to start of guide
            o = cal['out']  # distance from end of guide to other focal point
            foci = i + length + o
            if 'midpoint' in cal:
                w = cal['midpoint']
                cal['major'] = sqrt(foci * foci  + w * w) / 2
                cal['minor'] = w / 2
            else:
                entry = 'entry' in cal
                w, t, b = (cal['entry'], o, i) if entry else (cal['exit'], i ,o)
                t += length
                b = sqrt(b * b + w * w / 4) + sqrt(t * t + w * w / 4)
                cal['major'] = b / 2
                cal['minor'] = sqrt(b * b - foci * foci) / 2
            cal['offset'] = foci / 2 - i
        major = cal['major']
        minor = cal['minor']
        offset = cal['offset']
        return cls(
            major=major,
            minor=minor,
            offset=offset
        )

    def mccode_pars(self, post):
        p = {
            f'majorAxis{post}': self.major.to(unit='m').value,
            f'minorAxis{post}': self.minor.to(unit='m').value,
            f'majorAxisoffset{post}': self.offset.to(unit='m').value,
        }
        return p


class EllipticGuide(Guide):
    horizontal: PartialEllipse
    vertical:  PartialEllipse

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import sum, vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='deg')))
        length = cal['length']
        m = float_or_tuple_float(cal.get('m', 1.0))
        left = float_or_tuple_float(cal.get('left', m))
        right = float_or_tuple_float(cal.get('right', m))
        top = float_or_tuple_float(cal.get('top', m))
        bottom = float_or_tuple_float(cal.get('bottom', m))

        ms = left, right, top, bottom
        if any(isinstance(x, tuple) for x in ms):
            if not all(isinstance(x, tuple) for x in ms):
                raise ValueError('All m values should be scalars or tuples')
            count = len(left)
            if not all(len(x) == count for x in ms):
                raise ValueError('Tuple m values must be the same length')
            if not length.size == count:
                raise ValueError('Segment lengths and m values must be the same length')
            gl = sum(length)
        else:
            gl = length

        horizontal = PartialEllipse.from_calibration(gl, cal.get('horizontal'))
        vertical = PartialEllipse.from_calibration(gl, cal.get('vertical'))

        return cls(
            name=name,
            position=position,
            orientation=orientation,
            length=length,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
            horizontal=horizontal,
            vertical=vertical
        )

    def __mccode__(self) -> tuple[str, dict]:
        from scipp import sum
        p = dict()
        if isinstance(self.left, tuple):
            p['l'] = sum(self.length).to(unit='m').value
            p['nSegments'] = len(self.left)
            p['seglength'] = f'{self.name}_lens'
            for n in ('right', 'left', 'top', 'bottom'):
                p[f'mvalues{n}'] = f'{self.name}_{n}'
        else:
            p['l'] = self.length.to(unit='m').value
            for n in ('left', 'right', 'top', 'bottom'):
                p[f'm{n}'] = getattr(self, n)
        p.update(self.horizontal.mccode_pars('xw'))
        p.update(self.vertical.mccode_pars('yh'))

        return 'Elliptic_guide_gravity', p

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        if isinstance(self.left, tuple):
            assembler.declare_array('double', f'{self.name}_lens', self.length.to(unit='m').values)
            for n in ('left', 'right', 'top', 'bottom'):
                assembler.declare_array('double', f'{self.name}_{n}', getattr(self, n))
        super().to_mccode(assembler, at, rotate)



from scipp import Variable
from .component import Base


class Wire(Base):
    at: Variable
    to: Variable
    resistivity: Variable

    def extreme_path_corners(self, horizontal: Variable, vertical: Variable, unit=None):
        from ..spatial import combine_extremes
        from scipp import concat
        return combine_extremes([concat([self.at, self.to], 'vertices')], horizontal, vertical)

    def __eq__(self, other):
        if not isinstance(other, Wire):
            return False
        return self.at == other.at and self.to == other.to

    def approx(self, other):
        from scipp import allclose
        if not isinstance(other, Wire):
            return False
        return allclose(self.at, other.at) and allclose(self.to, other.to)

    def __post_init__(self):
        from scipp import DType
        if self.at.dtype != DType.vector3:
            raise RuntimeError("Wire starting point, at, must be a scipp.DType('vector3')")
        if self.to.dtype != DType.vector3:
            raise RuntimeError("Wire end point, to, must be a scipp.DType('vector3')")
        if self.to.unit != self.at.unit:
            raise RuntimeError("Wire end points must have the same unit")
        from ..utilities import is_scalar, has_compatible_unit
        if not is_scalar(self.resistivity):
            raise ValueError(f"The provided radius is not a scalar")
        if not has_compatible_unit(self.resistivity, 'Ohm/m'):
            raise ValueError(f"Provided unit {self.resistivity.unit} is not convertible to Ohm/m")

    @property
    def resistance(self) -> Variable:
        from scipp import sqrt, dot
        v = self.to - self.at
        return self.resistivity * sqrt(dot(v, v))

    @property
    def center_of_mass(self) -> Variable:
        return (self.at + self.to)/2

    def charge_position(self, a, b) -> Variable:
        if a+b == 0:
            raise RuntimeError("Sum of a and b must not zero")
        return self._charge_position(a/(a+b), b/(a+b))

    def _charge_position(self, fb, fa) -> Variable:
        if fb < 0 or fb > 1 or fa < 0 or fa > 1:
            raise RuntimeError("Both a and b should have the same sign")
        return fa * self.at + fb * self.to

    def continuous_position(self, fa) -> Variable:
        if fa < 0 or fa > 1:
            raise RuntimeError("Relative distance from at must be between 0 and 1")
        return (1 - fa) * self.at + fa * self.to

    def mcstas_parameters(self):
        from numpy import hstack
        center = (self.at + self.to)/2
        axis = (self.to - self.at)/2
        return hstack((center.value, axis.value))

    def center(self):
        return (self.at + self.to) / 2

    def end(self):
        return (self.to - self.at) / 2


class DiscreteWire(Wire):
    elements: int

    def __eq__(self, other):
        if not isinstance(other, DiscreteWire):
            return False
        return self.elements == other.elements and super().__eq__(other)

    def approx(self, other):
        from numpy import isclose
        if not isinstance(other, DiscreteWire):
            return False
        return isclose(self.elements, other.elements) and super().approx(other)

    def __post_init__(self):
        if self.elements == 0:
            raise RuntimeError("A discrete detector should have a finite number of elements")
        if self.elements < 0:
            raise RuntimeError("A discrete detector must have a positive number of elements")

    def charge_index(self, a, b) -> int:
        from math import floor
        if a * b < 0:
            raise RuntimeError("Both a and b should have the same sign")
        if a+b == 0:
            raise RuntimeError("Sum of a and b must not be zero")
        # Pick the *bin* index (0,N-1) -- but a/(a+b) is on the interval (0,1)
        n = floor(self.elements * (a / (a + b))) if b else self.elements - 1
        if 0 > n or n > self.elements:
            raise RuntimeError("Error calculating index")
        return n

    def index_position(self, index) -> Variable:
        # take the bin center as its position -- this is (n+1/2)/N along the whole length
        fb = (index + 0.5) / self.elements
        return super()._charge_position(1 - fb, fb)

    def charge_position(self, a, b) -> Variable:
        return self.index_position(self.charge_index(a, b))


class DiscreteTube(DiscreteWire):
    radius: Variable

    def __post_init__(self):
        from ..utilities import is_scalar, has_compatible_unit
        if not is_scalar(self.radius):
            raise ValueError(f"The provided radius is not a scalar")
        if not has_compatible_unit(self.radius, 'm'):
            raise ValueError(f"The provided 'radius' has unit {self.radius.unit} which is not convertible to m")

    def triangulate(self, unit=None):
        from scipp import vector, sqrt, dot, isclose, cross, arange, concat, flatten
        from scipp.spatial import rotations_from_rotvecs
        if unit is None:
            unit = self.at.unit
        lvec = self.to.to(unit=unit) - self.at.to(unit=unit)
        ll = sqrt(dot(lvec, lvec))
        # *a* vector perpendicular to l
        p = cross(lvec, vector([1., 0, 0]) if isclose(lvec.fields.z, ll) else vector([0, 0, 1.]))
        p = p/sqrt(dot(p, p)) * self.radius.to(unit=unit)

        a = arange(start=0, stop=360, step=30, dim='ring', unit='degree')
        r = rotations_from_rotvecs(a*lvec/ll)

        nvr = len(a)  # the number of vertices per ring
        ring = r * p
        li = self.at.to(unit=unit) + arange(start=0, stop=self.elements+1, dim='length') * lvec / self.elements
        vertices = flatten(li + ring, to='vertices')  # the order in the addition is important for flatten
        # 0, elements*[0,nvr), elements*nvr + 1
        vertices = concat((self.at.to(unit=unit), vertices, self.to.to(unit=unit)), 'vertices')
        # bottom cap
        faces = [[0, i + 1, (i + 1) % nvr + 1] for i in range(nvr)]
        # between rings
        for j in range(self.elements):
          z = 1 + j*nvr
          rf = [[[z + i, z + (i + 1) % nvr, z + (i + 1) % nvr + nvr],
                 [z + i, z + (i + 1) % nvr + nvr, z + i + nvr]] for i in range(nvr)]
          faces.extend([triangle for triangles in rf for triangle in triangles])
        # top cap
        last = len(vertices) - 1
        top = [[last, last - i - 1, last - (i + 1) % nvr -1] for i in range(nvr)]
        faces.extend(top)
        return vertices, faces

    def bounding_box(self, basis: Variable, unit=None):
        from ..spatial import bounding_box
        v, _ = self.triangulate(unit=unit)
        return bounding_box(v, basis)

    def __eq__(self, other):
        if not isinstance(other, DiscreteTube):
            return False
        return self.radius == other.radius and super().__eq__(other)

    def approx(self, other):
        from numpy import isclose
        if not isinstance(other, DiscreteTube):
            return False
        return isclose(self.radius, other.radius) and super().approx(other)


class He3Tube(DiscreteTube):
    pressure: Variable

    def __post_init__(self):
        from ..utilities import is_scalar, has_compatible_unit
        if not is_scalar(self.pressure):
            raise ValueError(f"The provided pressure is not a scalar")
        if not has_compatible_unit(self.pressure, 'Pa'):
            raise ValueError(f"The provided 'pressure' has unit {self.pressure.unit} which is not convertible to Pa")

    def __eq__(self, other):
        if not isinstance(other, He3Tube):
            return False
        return self.pressure == other.pressure and super().__eq__(other)

    def approx(self, other):
        from numpy import isclose
        if not isinstance(other, He3Tube):
            return False
        return isclose(self.pressure, other.pressure) and super().approx(other)
